"""The remote.exporter module exports resources to the coordinator and makes
them available to other clients on the same coordinator"""

import argparse
import asyncio
import logging
import sys
import os
import os.path
import traceback
import signal
import shutil
import subprocess
from urllib.parse import urlsplit
import warnings
from pathlib import Path
from typing import Dict, Type
from socket import gethostname, getfqdn

import attr
import grpc

from .config import ResourceConfig
from .common import ResourceEntry, queue_as_aiter
from .generated import labgrid_coordinator_pb2, labgrid_coordinator_pb2_grpc
from ..util import get_free_port, labgrid_version


exports: Dict[str, Type[ResourceEntry]] = {}
reexec = False


class ExporterError(Exception):
    pass


class BrokenResourceError(ExporterError):
    pass


class UnknownResourceError(ExporterError):
    pass


class InvalidResourceRequestError(ExporterError):
    pass


def log_subprocess_kernel_stack(logger, child):
    if child.poll() is not None:  # nothing to check if no longer running
        return
    try:
        with open(f"/proc/{child.pid}/stack", "r") as f:
            stack = f.read()
            stack = stack.strip()
    except PermissionError:
        return
    logger.info("current kernel stack of %s is:\n%s", child.args, stack)


@attr.s(eq=False)
class ResourceExport(ResourceEntry):
    """Represents a local resource exported via a specific protocol.

    The ResourceEntry attributes contain the information for the client.
    """

    host = attr.ib(default=gethostname(), validator=attr.validators.instance_of(str))
    proxy = attr.ib(default=None)
    proxy_required = attr.ib(default=False)
    local = attr.ib(init=False)
    local_params = attr.ib(init=False)
    start_params = attr.ib(init=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.logger = logging.getLogger(f"ResourceExport({self.cls})")
        # move the params to local_params
        self.local_params = self.params.copy()
        for key in self.local_params:
            del self.params[key]
        self.start_params = None
        self._broken = None

    # if something criticial failed for an export, we can mark it as
    # permanently broken
    @property
    def broken(self):
        return self._broken

    @broken.setter
    def broken(self, reason):
        assert self._broken is None
        assert type(reason) == str
        assert reason
        self._broken = reason
        # By setting the acquired field, we block places from using this
        # resource. For now, when trying to acquire a place with a match for
        # this resource, we get 'resource is already in used by <broken>',
        # instead of an unspecific error.
        self.data["acquired"] = "<broken>"
        self.logger.error("marked as broken: %s", reason)

    def _get_start_params(self):
        return {}

    def _get_params(self):
        return {}

    def _start(self, start_params):
        """Start exporting the local resource"""
        pass

    def _stop(self, start_params):
        """Stop exporting the local resource"""
        pass

    def start(self):
        assert not self.broken
        start_params = self._get_start_params()
        try:
            self._start(start_params)
        except Exception as e:
            self.broken = "start failed"
            self.logger.exception("failed to start with %s", start_params)
            raise BrokenResourceError("Failed to start resource") from e
        self.start_params = start_params

    def stop(self):
        assert not self.broken
        try:
            self._stop(self.start_params)
        except Exception:
            self.broken = "stop failed"
            self.logger.exception("failed to stop with %s", self.start_params)
            raise
        self.start_params = None

    def poll(self):
        # poll and check for updated params/avail
        self.local.poll()

        if self.broken:
            pass  # don't touch broken resources
        elif self.local.avail and self.acquired:
            start_params = self._get_start_params()
            if self.start_params is None:
                self.start()
            elif self.start_params != start_params:
                self.logger.info("restart needed (%s -> %s)", self.start_params, start_params)
                self.stop()
                self.start()
        else:
            if self.start_params is not None:
                self.stop()

        # check if resulting information has changed
        dirty = False
        if self.avail != (self.local.avail and not self.broken):
            self.data["avail"] = self.local.avail and not self.broken
            dirty = True
        params = self._get_params()
        if not params.get("extra"):
            params["extra"] = {}
        params["extra"]["proxy_required"] = self.proxy_required
        params["extra"]["proxy"] = self.proxy
        if self.broken:
            params["extra"]["broken"] = self.broken
        if self.params != params:
            self.data["params"].update(params)
            dirty = True

        return dirty

    def acquire(self, *args, **kwargs):
        if self.broken:
            raise BrokenResourceError(f"cannot acquire broken resource (original reason): {self.broken}")
        super().acquire(*args, **kwargs)
        self.poll()

    def release(self, *args, **kwargs):
        if self.broken:
            raise BrokenResourceError(f"cannot release broken resource (original reason): {self.broken}")
        super().release(*args, **kwargs)
        self.poll()


@attr.s(eq=False)
class SerialPortExport(ResourceExport):
    """ResourceExport for a USB or Raw SerialPort"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.cls == "RawSerialPort":
            from ..resource.serialport import RawSerialPort

            self.local = RawSerialPort(target=None, name=None, **self.local_params)
        elif self.cls == "USBSerialPort":
            from ..resource.udev import USBSerialPort

            self.local = USBSerialPort(target=None, name=None, **self.local_params)
        self.data["cls"] = "NetworkSerialPort"
        self.child = None
        self.port = None
        self.ser2net_bin = shutil.which("ser2net")
        if self.ser2net_bin is None:
            if os.path.isfile("/usr/sbin/ser2net"):
                self.ser2net_bin = "/usr/sbin/ser2net"

            if self.ser2net_bin is None:
                warnings.warn("ser2net binary not found, falling back to /usr/bin/ser2net")
                self.ser2net_bin = "/usr/bin/ser2net"

    def __del__(self):
        if self.child is not None:
            self.stop()

    def _get_start_params(self):
        return {
            "path": self.local.port,
        }

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "port": self.port,
            "speed": self.local.speed,
            "extra": {
                "path": self.local.port,
            },
        }

    def _start(self, start_params):
        """Start ``ser2net`` subprocess"""
        assert self.local.avail
        assert self.child is None
        assert start_params["path"].startswith("/dev/")
        self.port = get_free_port()

        # Ser2net has switched to using YAML format at version 4.0.0.
        result = subprocess.run([self.ser2net_bin, "-v"], capture_output=True, text=True)
        _, _, version = str(result.stdout).split(" ")
        version = tuple(map(int, version.strip().split(".")))

        # There is a bug in ser2net between 4.4.0 and 4.6.1 where it
        # returns 1 on a successful call to 'ser2net -v'. We don't want
        # a failure because of this, so raise an error only if ser2net
        # is not one of those versions.
        if version not in [(4, 4, 0), (4, 5, 0), (4, 5, 1), (4, 6, 0), (4, 6, 1)] and result.returncode == 1:
            raise ExporterError(f"ser2net {version} returned a nonzero code during version check.")

        if version >= (4, 2, 0):
            cmd = [
                self.ser2net_bin,
                "-d",
                "-n",
                "-Y",
                f"connection: &con01#  accepter: telnet(rfc2217,mode=server),tcp,{self.port}",
                "-Y",
                f"  connector: serialdev(nouucplock=true),{start_params['path']},{self.local.speed}n81,local",  # pylint: disable=line-too-long
                "-Y",
                "  options:",
                "-Y",
                "    max-connections: 10",
            ]
        else:
            cmd = [
                self.ser2net_bin,
                "-d",
                "-n",
                "-u",
                "-C",
                f"{self.port}:telnet:0:{start_params['path']}:{self.local.speed} NONE 8DATABITS 1STOPBIT LOCAL",  # pylint: disable=line-too-long
            ]
        self.logger.info("Starting ser2net with: %s", " ".join(cmd))
        self.child = subprocess.Popen(cmd)
        try:
            self.child.wait(timeout=0.5)
            raise ExporterError(f"ser2net for {start_params['path']} exited immediately")
        except subprocess.TimeoutExpired:
            # good, ser2net didn't exit immediately
            pass
        self.logger.info("started ser2net for %s on port %d", start_params["path"], self.port)

    def _stop(self, start_params):
        """Stop ``ser2net`` subprocess"""
        assert self.child
        child = self.child
        self.child = None
        port = self.port
        self.port = None
        child.terminate()
        try:
            child.wait(2.0)  # ser2net takes about a second to react
        except subprocess.TimeoutExpired:
            self.logger.warning("ser2net for %s still running after SIGTERM", start_params["path"])
            log_subprocess_kernel_stack(self.logger, child)
            child.kill()
            child.wait(1.0)
        self.logger.info("stopped ser2net for %s on port %d", start_params["path"], port)


exports["USBSerialPort"] = SerialPortExport
exports["RawSerialPort"] = SerialPortExport


@attr.s(eq=False)
class NetworkInterfaceExport(ResourceExport):
    """ResourceExport for a network interface"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.cls == "NetworkInterface":
            from ..resource.base import NetworkInterface

            self.local = NetworkInterface(target=None, name=None, **self.local_params)
        elif self.cls == "USBNetworkInterface":
            from ..resource.udev import USBNetworkInterface

            self.local = USBNetworkInterface(target=None, name=None, **self.local_params)
        self.data["cls"] = "RemoteNetworkInterface"

    def _get_params(self):
        """Helper function to return parameters"""
        params = {
            "host": self.host,
            "ifname": self.local.ifname,
        }
        if self.cls == "USBNetworkInterface":
            params["extra"] = {
                "state": self.local.if_state,
            }

        return params


exports["USBNetworkInterface"] = NetworkInterfaceExport
exports["NetworkInterface"] = NetworkInterfaceExport


@attr.s(eq=False)
class USBGenericExport(ResourceExport):
    """ResourceExport for USB devices accessed directly from userspace"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        local_cls_name = self.cls
        self.data["cls"] = f"Network{self.cls}"
        from ..resource import udev

        local_cls = getattr(udev, local_cls_name)
        self.local = local_cls(target=None, name=None, **self.local_params)

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
        }


@attr.s(eq=False)
class USBSigrokExport(USBGenericExport):
    """ResourceExport for USB devices accessed directly from userspace"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "driver": self.local.driver,
            "channels": self.local.channels,
            "channel_group": self.local.channel_group,
        }


@attr.s(eq=False)
class USBSDMuxExport(USBGenericExport):
    """ResourceExport for USB devices accessed directly from userspace"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "control_path": self.local.control_path,
        }


@attr.s(eq=False)
class USBSDWireExport(USBGenericExport):
    """ResourceExport for USB devices accessed directly from userspace"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "control_serial": self.local.control_serial,
        }


@attr.s(eq=False)
class USBSDWire3Export(USBGenericExport):
    """ResourceExport for USB devices accessed directly from userspace"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "control_serial": self.local.control_serial,
        }


@attr.s(eq=False)
class USBAudioInputExport(USBGenericExport):
    """ResourceExport for ports on switchable USB hubs"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "index": self.local.index,
            "alsa_name": self.local.alsa_name,
        }


@attr.s(eq=False)
class SiSPMPowerPortExport(USBGenericExport):
    """ResourceExport for ports on GEMBRID switches"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "index": self.local.index,
        }


@attr.s(eq=False)
class USBPowerPortExport(USBGenericExport):
    """ResourceExport for ports on switchable USB hubs"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "index": self.local.index,
        }


@attr.s(eq=False)
class USBDeditecRelaisExport(USBGenericExport):
    """ResourceExport for outputs on deditec relais"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "index": self.local.index,
            "invert": self.local.invert,
        }


@attr.s(eq=False)
class USBHIDRelayExport(USBGenericExport):
    """ResourceExport for outputs on simple USB HID relays"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "busnum": self.local.busnum,
            "devnum": self.local.devnum,
            "path": self.local.path,
            "vendor_id": self.local.vendor_id,
            "model_id": self.local.model_id,
            "index": self.local.index,
            "invert": self.local.invert,
        }


@attr.s(eq=False)
class USBFlashableExport(USBGenericExport):
    """ResourceExport for Flashable USB devices"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _get_params(self):
        p = super()._get_params()
        p["devnode"] = self.local.devnode
        return p


@attr.s(eq=False)
class USBGenericRemoteExport(USBGenericExport):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.data["cls"] = f"Remote{self.cls}".replace("Network", "")


exports["AndroidFastboot"] = USBGenericExport
exports["AndroidUSBFastboot"] = USBGenericRemoteExport
exports["DFUDevice"] = USBGenericExport
exports["IMXUSBLoader"] = USBGenericExport
exports["MXSUSBLoader"] = USBGenericExport
exports["RKUSBLoader"] = USBGenericExport
exports["AlteraUSBBlaster"] = USBGenericExport
exports["SigrokUSBDevice"] = USBSigrokExport
exports["SigrokUSBSerialDevice"] = USBSigrokExport
exports["USBSDMuxDevice"] = USBSDMuxExport
exports["USBSDWireDevice"] = USBSDWireExport
exports["USBSDWire3Device"] = USBSDWire3Export
exports["USBDebugger"] = USBGenericExport
exports["USBHub"] = USBGenericRemoteExport
exports["USBMassStorage"] = USBGenericExport
exports["USBVideo"] = USBGenericExport
exports["USBAudioInput"] = USBAudioInputExport
exports["USBTMC"] = USBGenericExport
exports["SiSPMPowerPort"] = SiSPMPowerPortExport
exports["USBPowerPort"] = USBPowerPortExport
exports["DeditecRelais8"] = USBDeditecRelaisExport
exports["HIDRelay"] = USBHIDRelayExport
exports["USBFlashableDevice"] = USBFlashableExport
exports["LXAUSBMux"] = USBGenericExport


@attr.s(eq=False)
class ProviderGenericExport(ResourceExport):
    """ResourceExport for Resources derived from BaseProvider"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        local_cls_name = self.cls
        self.data["cls"] = f"Remote{self.cls}"
        from ..resource import provider

        local_cls = getattr(provider, local_cls_name)
        self.local = local_cls(target=None, name=None, **self.local_params)

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "internal": self.local.internal,
            "external": self.local.external,
        }


exports["TFTPProvider"] = ProviderGenericExport
exports["NFSProvider"] = ProviderGenericExport
exports["HTTPProvider"] = ProviderGenericExport


@attr.s
class EthernetPortExport(ResourceExport):
    """ResourceExport for a ethernet interface"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        from ..resource.ethernetport import SNMPEthernetPort

        self.data["cls"] = "EthernetPort"
        self.local = SNMPEthernetPort(target=None, name=None, **self.local_params)

    def _get_params(self):
        """Helper function to return parameters"""
        return {"switch": self.local.switch, "interface": self.local.interface, "extra": self.local.extra}


exports["SNMPEthernetPort"] = EthernetPortExport


@attr.s(eq=False)
class GPIOSysFSExport(ResourceExport):
    _gpio_sysfs_path_prefix = "/sys/class/gpio"

    """ResourceExport for GPIO lines accessed directly from userspace"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.cls == "SysfsGPIO":
            from ..resource.base import SysfsGPIO

            self.local = SysfsGPIO(target=None, name=None, **self.local_params)
        elif self.cls == "MatchedSysfsGPIO":
            from ..resource.udev import MatchedSysfsGPIO

            self.local = MatchedSysfsGPIO(target=None, name=None, **self.local_params)
        self.data["cls"] = "NetworkSysfsGPIO"
        self.export_path = Path(GPIOSysFSExport._gpio_sysfs_path_prefix, f"gpio{self.local.index}")
        self.system_exported = False

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            "host": self.host,
            "index": self.local.index,
        }

    def _get_start_params(self):
        return {
            "index": self.local.index,
        }

    def _start(self, start_params):
        """Start a GPIO export to userspace"""
        index = start_params["index"]

        if self.export_path.exists():
            self.system_exported = True
            return

        export_sysfs_path = os.path.join(GPIOSysFSExport._gpio_sysfs_path_prefix, "export")
        with open(export_sysfs_path, mode="wb") as export:
            export.write(str(index).encode("utf-8"))

    def _stop(self, start_params):
        """Disable a GPIO export to userspace"""
        index = start_params["index"]

        if self.system_exported:
            return

        export_sysfs_path = os.path.join(GPIOSysFSExport._gpio_sysfs_path_prefix, "unexport")
        with open(export_sysfs_path, mode="wb") as unexport:
            unexport.write(str(index).encode("utf-8"))


exports["SysfsGPIO"] = GPIOSysFSExport
exports["MatchedSysfsGPIO"] = GPIOSysFSExport


@attr.s
class NetworkServiceExport(ResourceExport):
    """ResourceExport for a NetworkService

    This checks if the address has a interface suffix and then provides the
    neccessary proxy information.
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        from ..resource.networkservice import NetworkService

        self.data["cls"] = "NetworkService"
        self.local = NetworkService(target=None, name=None, **self.local_params)
        if "%" in self.local_params["address"]:
            self.proxy_required = True

    def _get_params(self):
        """Helper function to return parameters"""
        return {
            **self.local_params,
        }


exports["NetworkService"] = NetworkServiceExport


@attr.s
class HTTPVideoStreamExport(ResourceExport):
    """ResourceExport for an HTTPVideoStream"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        from ..resource.httpvideostream import HTTPVideoStream

        self.data["cls"] = "HTTPVideoStream"
        self.local = HTTPVideoStream(target=None, name=None, **self.local_params)

    def _get_params(self):
        return self.local_params


exports["HTTPVideoStream"] = HTTPVideoStreamExport


@attr.s(eq=False)
class LXAIOBusNodeExport(ResourceExport):
    """ResourceExport for LXAIOBusNode devices accessed via the HTTP API"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        local_cls_name = self.cls
        self.data["cls"] = f"Network{self.cls}"
        from ..resource import lxaiobus

        local_cls = getattr(lxaiobus, local_cls_name)
        self.local = local_cls(target=None, name=None, **self.local_params)

    def _get_params(self):
        return self.local_params


exports["LXAIOBusPIO"] = LXAIOBusNodeExport


@attr.s(eq=False)
class AndroidNetFastbootExport(ResourceExport):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        local_cls_name = self.cls
        self.data["cls"] = f"Remote{self.cls}"
        from ..resource import fastboot

        local_cls = getattr(fastboot, local_cls_name)
        self.local = local_cls(target=None, name=None, **self.local_params)

    def _get_params(self):
        """Helper function to return parameters"""
        return {"host": self.host, **self.local_params}


exports["AndroidNetFastboot"] = AndroidNetFastbootExport


@attr.s(eq=False)
class YKUSHPowerPortExport(ResourceExport):
    """ResourceExport for YKUSHPowerPort devices"""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        local_cls_name = self.cls
        self.data["cls"] = f"Network{local_cls_name}"
        from ..resource import ykushpowerport

        local_cls = getattr(ykushpowerport, local_cls_name)
        self.local = local_cls(target=None, name=None, **self.local_params)

    def _get_params(self):
        return {"host": self.host, **self.local_params}


exports["YKUSHPowerPort"] = YKUSHPowerPortExport


class Exporter:
    def __init__(self, config) -> None:
        """Set up internal datastructures on successful connection:
        - Setup loop, name, authid and address
        - Join the coordinator as an exporter"""
        self.config = config
        self.loop = asyncio.get_running_loop()
        self.name = config["name"]
        self.hostname = config["hostname"]
        self.isolated = config["isolated"]

        # It seems since https://github.com/grpc/grpc/pull/34647, the
        # ping_timeout_ms default of 60 seconds overrides keepalive_timeout_ms,
        # so set it as well.
        # Use GRPC_VERBOSITY=DEBUG GRPC_TRACE=http_keepalive for debugging.
        channel_options = [
            ("grpc.keepalive_time_ms", 7500),  # 7.5 seconds
            ("grpc.keepalive_timeout_ms", 10000),  # 10 seconds
            ("grpc.http2.ping_timeout_ms", 10000),  # 10 seconds
            ("grpc.http2.max_pings_without_data", 0),  # no limit
        ]

        # default to port 20408 if not specified
        if urlsplit(f"//{config['coordinator']}").port is None:
            config["coordinator"] += ":20408"

        self.channel = grpc.aio.insecure_channel(
            target=config["coordinator"],
            options=channel_options,
        )
        self.stub = labgrid_coordinator_pb2_grpc.CoordinatorStub(self.channel)
        self.out_queue = asyncio.Queue()
        self.pump_task = None

        self.poll_task = None

        self.groups = {}

    async def run(self) -> None:
        self.pump_task = self.loop.create_task(self.message_pump())
        self.send_started()

        config_template_env = {
            "env": os.environ,
            "isolated": self.isolated,
            "hostname": self.hostname,
            "name": self.name,
        }
        resource_config = ResourceConfig(self.config["resources"], config_template_env)
        for group_name, group in resource_config.data.items():
            group_name = str(group_name)
            for resource_name, params in group.items():
                resource_name = str(resource_name)
                if resource_name == "location":
                    continue
                if params is None:
                    continue
                cls = params.pop("cls", resource_name)

                # this may call back to acquire the resource immediately
                await self.add_resource(group_name, resource_name, cls, params)

            # flush queued message
            while not self.pump_task.done():
                try:
                    await asyncio.wait_for(self.out_queue.join(), timeout=1)
                    break
                except asyncio.TimeoutError:
                    if self.pump_task.done():
                        await self.pump_task
                        logging.debug("pump task exited, shutting down exporter")
                        return

        logging.info("creating poll task")
        self.poll_task = self.loop.create_task(self.poll())

        (done, pending) = await asyncio.wait((self.pump_task, self.poll_task), return_when=asyncio.FIRST_COMPLETED)
        logging.debug("task(s) %s exited, shutting down exporter", done)
        for task in pending:
            task.cancel()

        try:
            await self.pump_task
            await self.poll_task
        except asyncio.CancelledError:
            return

    def send_started(self):
        msg = labgrid_coordinator_pb2.ExporterInMessage()
        msg.startup.version = labgrid_version()
        msg.startup.name = self.name
        self.out_queue.put_nowait(msg)

    async def message_pump(self):
        got_message = False
        try:
            async for out_message in self.stub.ExporterStream(queue_as_aiter(self.out_queue)):
                got_message = True
                logging.debug("received message %s", out_message)
                kind = out_message.WhichOneof("kind")
                if kind == "hello":
                    print("Exporter ready", flush=True)
                    logging.info("connected to coordinator version %s", out_message.hello.version)
                elif kind == "set_acquired_request":
                    logging.debug("acquire request")
                    success = False
                    reason = None
                    try:
                        if out_message.set_acquired_request.place_name:
                            await self.acquire(
                                out_message.set_acquired_request.group_name,
                                out_message.set_acquired_request.resource_name,
                                out_message.set_acquired_request.place_name,
                            )
                        else:
                            await self.release(
                                out_message.set_acquired_request.group_name,
                                out_message.set_acquired_request.resource_name,
                            )
                        success = True
                    except (BrokenResourceError, InvalidResourceRequestError, UnknownResourceError) as e:
                        reason = e.args[0]
                        logging.warning("set_acquired_request failed: %s", reason)
                    finally:
                        in_message = labgrid_coordinator_pb2.ExporterInMessage()
                        in_message.response.success = success
                        if reason:
                            in_message.response.reason = reason
                        logging.debug("queuing %s", in_message)
                        self.out_queue.put_nowait(in_message)
                        logging.debug("queued %s", in_message)
                else:
                    logging.debug("unknown request: %s", kind)
        except grpc.aio.AioRpcError as e:
            self.out_queue.put_nowait(None)  # let the sender side exit gracefully
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                if got_message:
                    logging.error("coordinator became unavailable: %s", e.details())
                else:
                    logging.error("coordinator is unavailable: %s", e.details())

                global reexec
                reexec = True
            else:
                logging.exception("unexpected grpc error in coordinator message pump task")
        except Exception:
            self.out_queue.put_nowait(None)  # let the sender side exit gracefully
            logging.exception("error in coordinator message pump")

            # only send command response when the other updates have left the queue
            # perhaps with queue join/task_done
            # this should be a command from the coordinator

    async def acquire(self, group_name, resource_name, place_name):
        resource = self.groups.get(group_name, {}).get(resource_name)
        if resource is None:
            raise UnknownResourceError(
                f"acquire request for unknown resource {group_name}/{resource_name} by {place_name}"
            )

        if resource.acquired:
            raise InvalidResourceRequestError(
                f"Resource {group_name}/{resource_name} is already acquired by {resource.acquired}"
            )

        try:
            resource.acquire(place_name)
        finally:
            await self.update_resource(group_name, resource_name)

    async def release(self, group_name, resource_name):
        resource = self.groups.get(group_name, {}).get(resource_name)
        if resource is None:
            raise UnknownResourceError(f"release request for unknown resource {group_name}/{resource_name}")

        if not resource.acquired:
            raise InvalidResourceRequestError(f"Resource {group_name}/{resource_name} is not acquired")

        try:
            resource.release()
        finally:
            await self.update_resource(group_name, resource_name)

    async def _poll_step(self):
        for group_name, group in self.groups.items():
            for resource_name, resource in group.items():
                if not isinstance(resource, ResourceExport):
                    continue
                try:
                    changed = resource.poll()
                except Exception:  # pylint: disable=broad-except
                    print(f"Exception while polling {resource}", file=sys.stderr)
                    traceback.print_exc(file=sys.stderr)
                    continue
                if changed:
                    await self.update_resource(group_name, resource_name)
                else:
                    # let other tasks run, see https://github.com/python/asyncio/issues/284
                    await asyncio.sleep(0)

    async def poll(self):
        while True:
            try:
                await asyncio.sleep(0.25)
                await self._poll_step()
            except asyncio.CancelledError:
                break
            except Exception:  # pylint: disable=broad-except
                traceback.print_exc(file=sys.stderr)

    async def stop(self):
        if self.poll_task is not None:
            self.poll_task.cancel()

    async def add_resource(self, group_name, resource_name, cls, params):
        """Add a resource to the exporter and update status on the coordinator"""
        print(f"add resource {group_name}/{resource_name}: {cls}/{params}")
        group = self.groups.setdefault(group_name, {})
        assert resource_name not in group
        export_cls = exports.get(cls, ResourceEntry)
        config = {
            "avail": export_cls is ResourceEntry,
            "cls": cls,
            "params": params,
        }
        proxy_req = self.isolated
        if issubclass(export_cls, ResourceExport):
            res = group[resource_name] = export_cls(
                config, host=self.hostname, proxy=getfqdn(), proxy_required=proxy_req
            )
            res.poll()
        else:
            config["params"]["extra"] = {
                "proxy": getfqdn(),
                "proxy_required": proxy_req,
            }
            group[resource_name] = export_cls(config)
        await self.update_resource(group_name, resource_name)

    async def update_resource(self, group_name, resource_name):
        """Update status on the coordinator"""
        resource = self.groups[group_name][resource_name]
        msg = labgrid_coordinator_pb2.ExporterInMessage()
        msg.resource.CopyFrom(resource.as_pb2())
        msg.resource.path.group_name = group_name
        msg.resource.path.resource_name = resource_name
        self.out_queue.put_nowait(msg)
        logging.info("queued update for resource %s/%s", group_name, resource_name)


async def amain(config) -> bool:
    exporter = Exporter(config)

    if inspect:
        inspect.exporter = exporter

    def _stop():
        asyncio.ensure_future(exporter.stop())

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, _stop)
    loop.add_signal_handler(signal.SIGTERM, _stop)

    await exporter.run()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--coordinator",
        metavar="HOST:PORT",
        type=str,
        default=os.environ.get("LG_COORDINATOR", "127.0.0.1:20408"),
        help="coordinator host and port",
    )
    parser.add_argument(
        "-n",
        "--name",
        dest="name",
        type=str,
        default=None,
        help="public name of this exporter (defaults to the system hostname)",
    )
    parser.add_argument(
        "--hostname",
        dest="hostname",
        type=str,
        default=None,
        help="hostname (or IP) published for accessing resources (defaults to the system hostname)",
    )
    parser.add_argument(
        "--fqdn", action="store_true", default=False, help="Use fully qualified domain name as default for hostname"
    )
    parser.add_argument("-d", "--debug", action="store_true", default=False, help="enable debug mode")
    parser.add_argument(
        "-i",
        "--isolated",
        action="store_true",
        default=False,
        help="enable isolated mode (always request SSH forwards)",
    )
    parser.add_argument("--pystuck", action="store_true", help="enable pystuck")
    parser.add_argument(
        "--pystuck-port", metavar="PORT", type=int, default=6667, help="use a different pystuck port than 6667"
    )
    parser.add_argument("resources", metavar="RESOURCES", type=str, help="resource config file name")

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    config = {
        "name": args.name or gethostname(),
        "hostname": args.hostname or (getfqdn() if args.fqdn else gethostname()),
        "resources": args.resources,
        "coordinator": args.coordinator,
        "isolated": args.isolated,
    }

    print(f"exporter name: {config['name']}")
    print(f"exporter hostname: {config['hostname']}")
    print(f"resource config file: {config['resources']}")

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    global inspect
    if args.pystuck:
        from types import SimpleNamespace

        inspect = SimpleNamespace()
        inspect.loop = loop

        import pystuck

        pystuck.run_server(port=args.pystuck_port)
    else:
        inspect = None

    asyncio.run(amain(config), debug=bool(args.debug))

    if reexec:
        exit(100)


if __name__ == "__main__":
    main()



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class SigrokDevice(Resource):
    """The SigrokDevice describes an attached sigrok device with driver,
    channel mapping and channel group

    Args:
        driver (str): driver to use with sigrok
        channels (str): a sigrok channel mapping as described in the sigrok-cli man page
        channel_group (str): a sigrok channel group as described in the sigrok-cli man page
    """

    driver = attr.ib(default="demo")
    channels = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    channel_group = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))



"""Control NETGEAR Plus devices via HTTP.

Available switch models:
https://github.com/foxey/py-netgear-plus?tab=readme-ov-file#supported-and-tested-netgear-modelsproducts-and-firmware-versions

The password defaults to "P4ssword", but can be configured on a per-device basis like this:

NetworkPowerPort:
  model: poe_netgear_plus
  host: 'http://username_is_unused:AnotherP4ssword@192.168.0.239/'
  index: 7

Omitting the password defaults as described above.

NetworkPowerPort:
  model: poe_netgear_plus
  host: 'http://192.168.0.239/'
  index: 7

"""

from urllib.parse import urlparse

from py_netgear_plus import NetgearSwitchConnector

from ..exception import ExecutionError


def _get_hostname_and_password(url: str) -> tuple[str, str]:
    """Obtain credentials from url or default and return hostname and password.

    If no password is in the URL return "P4ssword", which fulfills  the minimal requirements from Netgear:
    - 8-20 characters
    - at least one upper case character
    - at least one lower case character
    - at least one number

    Args:
        url: A URL with an optional basic auth prefix.

    Returns:
        A tuple of the hostname, and the extracted or default password

    """
    parse_result = urlparse(url)
    if parse_result.scheme != "http":
        raise ExecutionError(f"URL must start with http://, found {parse_result.scheme} for {url}.")

    password = "P4ssword" if parse_result.password is None else parse_result.password

    return parse_result.hostname, password


def power_set(host: str, _port: int, index: int, value: bool) -> None:
    """Set the PoE output index based for a given host.

    Args:
        host: The netloc with optional password e.g. "192.168.0.239" or ":P4ssword@192.168.0.239"
        _port: As the webserver of the switch is always on port 80, this is ignored
        index: Zero based access to the switches network ports
        value: Whether the port should enable PoE output

    """
    index = int(index)
    netgear_port_number = index + 1

    (hostname, password) = _get_hostname_and_password(host)

    sw = NetgearSwitchConnector(hostname, password)
    sw.autodetect_model()
    try:
        sw.get_login_cookie()
        sw._get_switch_metadata()
        if value:
            sw.turn_on_poe_port(netgear_port_number)
        else:
            sw.turn_off_poe_port(netgear_port_number)
    finally:
        sw.delete_login_cookie()


def power_get(host: str, _port: int, index: int) -> bool:
    """Determine whether a given Port has PoE enabled.

    Args:
        host: The netloc with optional password e.g. "192.168.0.239" or ":P4ssword@192.168.0.239"
        _port: As the webserver of the switch is always on port 80, this is ignored
        index: Zero based access to the switches network ports

    Returns:
        Whether the PoE output is enabled.

    Raises:
        ExecutionError: In case the status dictionary contains unexpected PoE status values.

    """
    index = int(index)
    netgear_port_number = index + 1

    (hostname, password) = _get_hostname_and_password(host)

    sw = NetgearSwitchConnector(hostname, password)
    sw.autodetect_model()
    try:
        sw.get_login_cookie()
        sw._get_switch_metadata()
        data = sw._get_poe_port_config()
        key = f"port_{netgear_port_number}_poe_power_active"
        if data[key] == "on":
            return True
        if data[key] == "off":
            return False
        msg = f"Expected literal 'on'|'off', found {data[key]}"
        raise ExecutionError(msg)
    finally:
        sw.delete_login_cookie()



"""
UBUS jsonrpc interface for PoE management on OpenWrt devices. This comes in
handy if devices are connected to a PoE switch running OpenWrt.

The URL given in hosts in exporter.yaml must accept unauthenticated UBUS
calls for the two ``poe`` calls ``info`` and ``manage``.

Further information is availbe at https://openwrt.org/docs/techref/ubus#acls

NetworkPowerPort:
    model: ubus
    host: 'http://192.168.1.1/ubus'
    index: 1
"""

import requests


def jsonrpc_call(host, path, method, message):
    r = requests.post(
        host,
        json={
            "jsonrpc": "2.0",
            "id": 1,
            "method": "call",
            "params": ["00000000000000000000000000000000", path, method, message],
        },
    )
    r.raise_for_status()
    return r.json()["result"]


def power_set(host, port, index, value):
    assert port is None

    jsonrpc_call(host, "poe", "manage", {"port": f"lan{index}", "enable": value == 1})


def power_get(host, port, index):
    assert port is None

    poe_info = jsonrpc_call(host, "poe", "info", {})[1]

    assert f"lan{index}" in poe_info["ports"], f"Port lan{index} not found in {poe_info['ports']}"

    return poe_info["ports"][f"lan{index}"] != "Disabled"



import os
import socket
import subprocess

import pytest
from py.path import local

import labgrid.util.agentwrapper
from labgrid.util.agentwrapper import AgentError, AgentException, AgentWrapper
from labgrid.util.agent import b2s, s2b, py2s, s2py

@pytest.fixture(scope='function')
def subprocess_mock(mocker):
    import subprocess

    original = subprocess.Popen

    agent = local(labgrid.util.agentwrapper.__file__).dirpath('agent.py')

    def run(args, **kwargs):
        assert args[0] in ['rsync', 'ssh']
        if args[0] == 'rsync':
            src = local(args[-2])
            assert src == agent
            dst = args[-1]
            assert ':' in dst
            dst = dst.split(':', 1)[1]
            assert '/' not in dst
            assert dst.startswith('.labgrid_agent')
            return original(['true'], **kwargs)
        elif args[0] == 'ssh':
            assert '--' in args
            args = args[args.index('--')+1:]
            assert len(args) == 2
            assert args[0] == 'python3'
            assert args[1].startswith('.labgrid_agent')
            # we need to use the original here to get the coverage right
            return original(['python3', str(agent)], **kwargs)

    mocker.patch('subprocess.Popen', run)

def test_create(subprocess_mock):
    aw = AgentWrapper('localhost')
    aw.close()

def test_call(subprocess_mock):
    aw = AgentWrapper('localhost')
    assert aw.call('test') == []
    assert aw.call('test', 0) == [0]
    assert aw.call('test', 0, 1) == [1, 0]
    assert aw.call('test', 'foo') == ['foo']
    assert aw.call('test', '{') == ['{']

def test_proxy(subprocess_mock):
    aw = AgentWrapper('localhost')
    assert aw.test() == []
    assert aw.test( 0, 1) == [1, 0]

def test_bytes(subprocess_mock):
    aw = AgentWrapper('localhost')
    assert s2b(aw.test(b2s(b'\x00foo'))[0]) == b'\x00foo'

def test_pydata(subprocess_mock):
    with AgentWrapper('localhost') as aw:
        d = s2py(aw.test(py2s((1, 'foo', True)))[0])
        assert isinstance(d, tuple)
        assert d == (1, 'foo', True)

def test_exception(subprocess_mock):
    aw = AgentWrapper('localhost')

    with pytest.raises(AgentException) as excinfo:
        aw.error('foo')
    assert excinfo.value.args == ("ValueError('foo')",)

    with pytest.raises(AgentException) as excinfo:
        aw.error('bar')
    assert excinfo.value.args == ("ValueError('bar')",)

def test_error(subprocess_mock):
    aw = AgentWrapper('localhost')
    aw.agent.stdin.write(b'\x00')
    with pytest.raises(AgentError):
        aw.test()

def test_module(subprocess_mock):
    aw = AgentWrapper('localhost')
    dummy = aw.load('dummy')
    assert dummy.neg(1) == -1

def test_local():
    aw = AgentWrapper(None)

    assert aw.test() == []
    assert aw.test( 0, 1) == [1, 0]

    assert s2b(aw.test(b2s(b'\x00foo'))[0]) == b'\x00foo'

    with pytest.raises(AgentException) as excinfo:
        aw.error('foo')
    assert excinfo.value.args == ("ValueError('foo')",)

    with pytest.raises(AgentException) as excinfo:
        aw.error('bar')
    assert excinfo.value.args == ("ValueError('bar')",)

    dummy = aw.load('dummy')
    assert dummy.neg(1) == -1


def test_local_fdpass():
    aw = AgentWrapper(None)

    result = aw.test_fd()
    assert isinstance(result, tuple)
    assert result[0] == "dummy"
    assert isinstance(result[1], int)

    with os.fdopen(result[1]) as f:
        fdpath = os.readlink(f"/proc/self/fd/{f.fileno()}")
        assert fdpath.startswith("/memfd:test_fd")


@pytest.mark.parametrize('module_name', [
    'deditec_relais8',
    'sysfsgpio',
    'usb_hid_relay'
])
def test_all_modules(module_name: str) -> None:
    aw = AgentWrapper(None)

    aw.load(module_name)
    methods = aw.list()
    assert f'{module_name}.set' in methods
    assert f'{module_name}.get' in methods

def test_import_modules():
    import labgrid.util.agents
    import labgrid.util.agents.dummy
    from labgrid.util.agents import deditec_relais8, sysfsgpio



import os
import copy
import logging
import pytest

from .. import Environment
from ..consoleloggingreporter import ConsoleLoggingReporter
from ..util.helper import processwrapper
from ..logging import StepFormatter, StepLogger
from ..exceptions import NoStrategyFoundError

LABGRID_ENV_KEY = pytest.StashKey[Environment]()


@pytest.hookimpl(tryfirst=True)
def pytest_cmdline_main(config):
    def set_cli_log_level(level):
        nonlocal config

        try:
            current_level = config.getoption("log_cli_level") or config.getini("log_cli_level")
        except ValueError:
            return
        print(f"current_level: {current_level}")

        if isinstance(current_level, str):
            try:
                current_level = int(logging.getLevelName(current_level))
            except ValueError:
                current_level = None

        # If no level was set previously (via ini or cli) or current_level is
        # less verbose than level, set to new level.
        if current_level is None or level < current_level:
            config.option.log_cli_level = str(level)

    verbosity = config.getoption("verbose")
    if verbosity > 3: # enable with -vvvv
        set_cli_log_level(logging.DEBUG)
    elif verbosity > 2: # enable with -vvv
        set_cli_log_level(logging.CONSOLE)
    elif verbosity > 1: # enable with -vv
        set_cli_log_level(logging.INFO)


def configure_pytest_logging(config, plugin):
    if hasattr(plugin.log_cli_handler.formatter, "add_color_level"):
        plugin.log_cli_handler.formatter.add_color_level(logging.CONSOLE, "blue")
    plugin.log_cli_handler.setFormatter(StepFormatter(
        color=config.option.lg_colored_steps,
        parent=plugin.log_cli_handler.formatter,
    ))
    plugin.log_file_handler.setFormatter(StepFormatter(
        parent=plugin.log_file_handler.formatter,
    ))

    # Might be the same formatter instance, so get a reference for both before
    # changing either
    report_formatter = plugin.report_handler.formatter
    caplog_formatter = plugin.caplog_handler.formatter

    plugin.report_handler.setFormatter(StepFormatter(parent=report_formatter))
    plugin.report_handler.setFormatter(StepFormatter(parent=caplog_formatter))

@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    StepLogger.start()
    config.add_cleanup(StepLogger.stop)

    logging_plugin = config.pluginmanager.getplugin('logging-plugin')
    if logging_plugin:
        configure_pytest_logging(config, logging_plugin)

    config.addinivalue_line("markers",
                            "lg_feature: skip tests on envs/targets without given labgrid feature flags")
    config.addinivalue_line("markers",
                            "lg_xfail_feature: mark tests xfail on envs/targets with given labgrid feature flag")

    lg_log = config.option.lg_log
    if lg_log:
        ConsoleLoggingReporter(lg_log)
    lg_env = config.option.lg_env
    lg_coordinator = config.option.lg_coordinator

    env = None
    if lg_env is None:
        lg_env = os.environ.get('LG_ENV')
    if lg_env is not None:
        env = Environment(config_file=lg_env)
        if lg_coordinator is not None:
            env.config.set_option('coordinator_address', lg_coordinator)
    config.stash[LABGRID_ENV_KEY] = env

    processwrapper.enable_logging()

@pytest.hookimpl()
def pytest_collection_modifyitems(config, items):
    """This function matches function feature flags with those found in the
    environment and disables the item if no match is found"""
    env = config.stash[LABGRID_ENV_KEY]

    if not env:
        return

    have_feature = env.get_features() | env.get_target_features()

    for item in items:
        # pytest.mark.lg_feature
        lg_feature_signature = "pytest.mark.lg_feature(features: str | list[str])"
        want_feature = set()

        for marker in item.iter_markers("lg_feature"):
            if len(marker.args) != 1 or marker.kwargs:
                raise pytest.UsageError(f"Unexpected number of args/kwargs for {lg_feature_signature}")
            elif isinstance(marker.args[0], str):
                want_feature.add(marker.args[0])
            elif isinstance(marker.args[0], list):
                want_feature.update(marker.args[0])
            else:
                raise pytest.UsageError(f"Unsupported 'features' argument type ({type(marker.args[0])}) for {lg_feature_signature}")

        missing_feature = want_feature - have_feature
        if missing_feature:
            reason = f'unsupported feature(s): {", ".join(missing_feature)}'
            item.add_marker(pytest.mark.skip(reason=reason))

        # pytest.mark.lg_xfail_feature
        lg_xfail_feature_signature = "pytest.mark.lg_xfail_feature(feature: str, *, **xfail_kwargs), xfail_kwargs as pytest.mark.xfail expects them"
        for marker in item.iter_markers("lg_xfail_feature"):
            if len(marker.args) != 1:
                raise pytest.UsageError(f"Unexpected number of arguments for {lg_xfail_feature_signature}")
            elif not isinstance(marker.args[0], str):
                raise pytest.UsageError(f"Unsupported 'feature' argument type {type(marker.args[0])} for {lg_xfail_feature_signature}")
            if "condition" in marker.kwargs:
                raise pytest.UsageError(f"Unsupported 'condition' argument for {lg_xfail_feature_signature}")

            kwargs = copy.copy(marker.kwargs)
            reason = kwargs.pop("reason", marker.args[0])
            item.add_marker(
                pytest.mark.xfail(
                    condition=marker.args[0] in have_feature,
                    reason=reason,
                    **kwargs,
                )
            )

@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    """
    Skip test if one of the targets uses a strategy considered broken.
    """
    # Before any fixtures run for the test, check if the session-scoped strategy fixture was
    # requested (might have been executed already for a prior test). If that's the case and the
    # strategy is broken, skip the test.
    if "strategy" in item.fixturenames:
        env = item.config.stash[LABGRID_ENV_KEY]
        # skip test even if only one of the targets in the env has a broken strategy
        for target_name in env.config.get_targets():
            target = env.get_target(target_name)
            try:
                strategy = target.get_strategy()
                if strategy.broken:
                    pytest.skip(f"{strategy.__class__.__name__} is in broken state")
            except NoStrategyFoundError:
                pass



from itertools import chain
import attr

from ..factory import target_factory
from ..protocol import BootstrapProtocol
from ..step import step
from ..util.managedfile import ManagedFile
from ..util.helper import processwrapper
from .common import Driver


@target_factory.reg_driver
@attr.s(eq=False)
class OpenOCDDriver(Driver, BootstrapProtocol):
    bindings = {
        "interface": {
            "AlteraUSBBlaster",
            "NetworkAlteraUSBBlaster",
            "USBDebugger",
            "NetworkUSBDebugger",
        },
    }

    config = attr.ib(
        default=attr.Factory(list), validator=attr.validators.optional(attr.validators.instance_of((str, list)))
    )
    search = attr.ib(
        default=attr.Factory(list), validator=attr.validators.optional(attr.validators.instance_of((str, list)))
    )
    image = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    interface_config = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    board_config = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    load_commands = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(list)))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool("openocd")
            self.config = self.target.env.config.resolve_path_str_or_list(self.config)
            self.search = self.target.env.config.resolve_path_str_or_list(self.search)
        else:
            self.tool = "openocd"
            if isinstance(self.config, str):
                self.config = [self.config]
            if isinstance(self.search, str):
                self.search = [self.search]

    def _get_usb_path_cmd(self):
        # OpenOCD supports "adapter usb location" since a1b308ab, released with 0.11.0-rc1
        return ["--command", f'adapter usb location "{self.interface.path}"']

    def _run_commands(self, commands: list):
        cmd = [self.tool]
        cmd += chain.from_iterable(("--search", path) for path in self.search)
        cmd += self._get_usb_path_cmd()

        managed_configs = []
        for config in self.config:
            mconfig = ManagedFile(config, self.interface)
            mconfig.sync_to_resource()
            managed_configs.append(mconfig)

        if self.interface_config:
            cmd.append("--file")
            cmd.append(f"interface/{self.interface_config}")

        if self.board_config:
            cmd.append("--file")
            cmd.append(f"board/{self.board_config}")

        for mconfig in managed_configs:
            cmd.append("--file")
            cmd.append(mconfig.get_remote_path())

        cmd += chain.from_iterable(("--command", f"{command}") for command in commands)
        processwrapper.check_output(self.interface.wrap_command(cmd), print_on_silent_log=True)

    @Driver.check_active
    @step(args=["filename"])
    def load(self, filename=None):
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        mf = ManagedFile(filename, self.interface)
        mf.sync_to_resource()

        if self.load_commands is None:
            commands = [
                "init",
                f"bootstrap {mf.get_remote_path()}",
                "shutdown",
            ]
        else:
            commands = [c.format(filename=mf.get_remote_path()) for c in self.load_commands]

        self._run_commands(commands)

    @Driver.check_active
    @step(args=["commands"])
    def execute(self, commands: list):
        self._run_commands(commands)



def test_shell(command):
    stdout, stderr, returncode = command.run("cat /proc/version")
    assert returncode == 0
    assert len(stdout) > 0
    assert len(stderr) == 0
    assert "Linux" in stdout[0]

    stdout, stderr, returncode = command.run("false")
    assert returncode != 0
    assert len(stdout) == 0
    assert len(stderr) == 0



import attr

from ..factory import target_factory
from .common import NetworkResource, Resource


@target_factory.reg_resource
@attr.s(eq=False)
class DediprogFlasher(Resource):
    vcc = attr.ib(validator=attr.validators.in_(("3.5V", "2.5V", "1.8V")))


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkDediprogFlasher(NetworkResource):
    vcc = attr.ib(validator=attr.validators.in_(("3.5V", "2.5V", "1.8V")))



import re
import requests

PORT = 80


def power_set(host, port, index, value):
    index = int(index)
    assert 1 <= index <= 4
    # access the web interface...
    if value:
        portstring = {1: "1uuu", 2: "u1uu", 3: "uu1u", 4: "uuu1"}
    else:
        portstring = {1: "0uuu", 2: "u0uu", 3: "uu0u", 4: "uuu0"}
    r = requests.get(f"http://{host}:{port}/tgi/control.tgi?l=p:admin:admin&p={portstring[index]}")
    r.raise_for_status()


def power_get(host, port, index):
    index = int(index)
    assert 1 <= index <= 4
    # get the contents of the main page
    r = requests.get(f"http://{host}:{port}/tgi/control.tgi?l=p:admin:admin&p=l")
    r.raise_for_status()
    m = re.match(r".*(\d) (\d) (\d) (\d).*", r.text)
    states = {"0": False, "1": True}
    value = m.group(index)
    return states[value]



import shlex
import time
import math
from importlib import import_module

import attr

from ..exceptions import InvalidConfigError
from ..factory import target_factory
from ..protocol import PowerProtocol, DigitalOutputProtocol, ResetProtocol
from ..resource import NetworkPowerPort
from ..step import step
from ..util.proxy import proxymanager
from ..util.helper import processwrapper
from .common import Driver
from .exception import ExecutionError


@attr.s(eq=False)
class PowerResetMixin(ResetProtocol):
    """
    ResetMixin implements the ResetProtocol for drivers which support the PowerProtocol
    """
    priorities = {ResetProtocol: -10}

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @Driver.check_active
    @step()
    def reset(self):
        self.cycle()

@target_factory.reg_driver
@attr.s(eq=False)
class ManualPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """ManualPowerDriver - Driver to tell the user to control a target's power"""

    @Driver.check_active
    @step()
    def on(self):
        self.target.interact(
            f"Turn the target {self.target.name} ON and press enter"
        )

    @Driver.check_active
    @step()
    def off(self):
        self.target.interact(
            f"Turn the target {self.target.name} OFF and press enter"
        )

    @Driver.check_active
    @step()
    def cycle(self):
        self.target.interact(
            f"CYCLE the target {self.target.name} and press enter"
        )


@target_factory.reg_driver
@attr.s(eq=False)
class SiSPMPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """SiSPMPowerDriver - Driver using a SiS-PM (Silver Shield PM) to control a
       target's power using the sispmctl tool - http://sispmctl.sourceforge.net/"""

    bindings = {"port": {"SiSPMPowerPort", "NetworkSiSPMPowerPort"}, }
    delay = attr.ib(default=2.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool('sispmctl')
        else:
            self.tool = 'sispmctl'

    def _get_sispmctl_prefix(self):
        return self.port.command_prefix + [
            self.tool,
            "-U", f"{self.port.busnum:03d}:{self.port.devnum:03d}",
        ]

    @Driver.check_active
    @step()
    def on(self):
        cmd = ['-o', str(self.port.index)]
        processwrapper.check_output(self._get_sispmctl_prefix() + cmd)

    @Driver.check_active
    @step()
    def off(self):
        cmd = ['-f', str(self.port.index)]
        processwrapper.check_output(self._get_sispmctl_prefix() + cmd)

    @Driver.check_active
    @step()
    def cycle(self):
        self.off()
        time.sleep(self.delay)
        self.on()

    @Driver.check_active
    @step()
    def get(self):
        cmd = ['-q', '-g', str(self.port.index)]
        output = processwrapper.check_output(self._get_sispmctl_prefix() + cmd)
        if output.strip() == b"on":
            return True
        if output.strip() == b"off":
            return False
        raise ExecutionError(f"Did not find port status in sispmctl output ({repr(output)})")


@target_factory.reg_driver
@attr.s(eq=False)
class ExternalPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """ExternalPowerDriver - Driver using an external command to control a target's power"""
    cmd_on = attr.ib(validator=attr.validators.instance_of(str))
    cmd_off = attr.ib(validator=attr.validators.instance_of(str))
    cmd_cycle = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    delay = attr.ib(default=2.0, validator=attr.validators.instance_of(float))

    @Driver.check_active
    @step()
    def on(self):
        cmd = shlex.split(self.cmd_on)
        processwrapper.check_output(cmd)

    @Driver.check_active
    @step()
    def off(self):
        cmd = shlex.split(self.cmd_off)
        processwrapper.check_output(cmd)

    @Driver.check_active
    @step()
    def cycle(self):
        if self.cmd_cycle is not None:
            cmd = shlex.split(self.cmd_cycle)
            processwrapper.check_output(cmd)
        else:
            self.off()
            time.sleep(self.delay)
            self.on()

@target_factory.reg_driver
@attr.s(eq=False)
class NetworkPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """NetworkPowerDriver - Driver using a networked power switch to control a target's power"""
    bindings = {"port": NetworkPowerPort, }
    delay = attr.ib(default=2.0, validator=attr.validators.instance_of(float))


    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # TODO: allow backends to register models with other names
        self.backend = import_module(
            f".power.{self.port.model}",
            __package__
        )
        self._host = None
        self._port = None

    def on_activate(self):
        # we can only forward if the backend knows which port to use
        backend_port = getattr(self.backend, 'PORT', None)
        if backend_port:
            self._host, self._port = proxymanager.get_host_and_port(
                self.port, force_port=backend_port
            )
        else:
            # try handling the host parameter as a URL and extract required data from it
            if not self.set_proxy_from_url():
                # fallback
                self._host = self.port.host
                self._port = None

    def set_proxy_from_url(self):
        """
        Some power backends (e.g. rest, simplerest) expect a URL in the host parameter. Try to
        handle the host parameter as a URL, extract the power backend port from it and set new
        proxied host parameter.
        """
        from urllib.parse import urlparse

        url = urlparse(self.port.host)
        if not url.hostname:
            return False

        if url.port:
            backend_port = url.port
        elif url.scheme == 'http':
            backend_port = 80
        elif url.scheme == 'https':
            backend_port = 443
        else:
            # unknown scheme specifier
            return False

        self._host, self._port = proxymanager.get_host_and_port(
            self.port, force_port=backend_port
        )

        # construct proxied host parameter
        query = f'?{url.query}' if url.query else ''
        user_pass = f'{url.netloc.split("@")[0]}@' if '@' in url.netloc else ''
        self._host = f'{url.scheme}://{user_pass}{self._host}:{self._port}{url.path}{query}'
        self._port = None

        return True

    @Driver.check_active
    @step()
    def on(self):
        self.backend.power_set(self._host, self._port, self.port.index, True)

    @Driver.check_active
    @step()
    def off(self):
        self.backend.power_set(self._host, self._port, self.port.index, False)

    @Driver.check_active
    @step()
    def cycle(self):
        self.off()
        time.sleep(self.delay)
        self.on()

    @Driver.check_active
    def get(self):
        return self.backend.power_get(self._host, self._port, self.port.index)

@target_factory.reg_driver
@attr.s(eq=False)
class DigitalOutputPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """
    DigitalOutputPowerDriver uses a DigitalOutput to control the power
    of a DUT.
    """
    bindings = {"output": DigitalOutputProtocol, }
    delay = attr.ib(default=1.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @Driver.check_active
    @step()
    def on(self):
        self.output.set(True)

    @Driver.check_active
    @step()
    def off(self):
        self.output.set(False)

    @Driver.check_active
    @step()
    def cycle(self):
        self.off()
        time.sleep(self.delay)
        self.on()

    @Driver.check_active
    @step()
    def get(self):
        return self.output.get()

@target_factory.reg_driver
@attr.s(eq=False)
class YKUSHPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """YKUSHPowerDriver - Driver using a YEPKIT YKUSH switchable USB hub
        to control a target's power - https://www.yepkit.com/products/ykush.
        Uses ykushcmd tool to control the hub."""
    bindings = {"port": {"YKUSHPowerPort", "NetworkYKUSHPowerPort"} }
    delay = attr.ib(default=2.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool('ykushcmd') or 'ykushcmd'
        else:
            self.tool = 'ykushcmd'

    def on_activate(self):
        # Search for the model of the YKUSH device
        for model in ["ykush", "ykushxs", "ykush3"]:
            cmd = self.port.command_prefix + [
                self.tool,
                model,
                "-l"
            ]
            output = processwrapper.check_output(cmd)
            if self.port.serial in output.decode("utf-8"):
                self.model = model
                break
        else:
            raise InvalidConfigError(f"Could not find YKUSH device with serial {self.port.serial}")

    @Driver.check_active
    @step()
    def on(self):
        cmd = [
            self.tool,
            self.model,
            "-s", f"{self.port.serial}",
            "-u", f"{self.port.index}"
        ]
        processwrapper.check_output(self.port.command_prefix + cmd)

    @Driver.check_active
    @step()
    def off(self):
        cmd = [
            self.tool,
            self.model,
            "-s", f"{self.port.serial}",
            "-d", f"{self.port.index}"
        ]
        processwrapper.check_output(self.port.command_prefix + cmd)

    @Driver.check_active
    @step()
    def cycle(self):
        self.off()
        time.sleep(self.delay)
        self.on()

    @Driver.check_active
    def get(self):
        cmd = [
            self.tool,
            self.model,
            "-s", f"{self.port.serial}",
            "-g", f"{self.port.index}"
        ]
        res = processwrapper.check_output(self.port.command_prefix + cmd)
        res = res.decode("utf-8")

        # the example of ykushcmd -g like below:
        # cmd: ykushcmd -g 1
        # output: Downstream port 1 is ON/OFF
        check_str = "Downstream port {port} is {status}"
        if check_str.format(port=self.port.index, status="ON") in res:
            return True
        if check_str.format(port=self.port.index, status="OFF") in res:
            return False

        raise ExecutionError(f"Could not find port string in ykushcmd output: {res}")


@target_factory.reg_driver
@attr.s(eq=False)
class USBPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """USBPowerDriver - Driver using a power switchable USB hub and the uhubctl
    tool (https://github.com/mvp/uhubctl) to control a target's power"""

    bindings = {"hub": {"USBPowerPort", "NetworkUSBPowerPort"}, }
    delay = attr.ib(default=2.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool('uhubctl')
        else:
            self.tool = 'uhubctl'

    def _switch(self, cmd):
        cmd = self.hub.command_prefix + [
            self.tool,
            "-l", self.hub.path,
            "-p", str(self.hub.index),
            "-r", "100", # use 100 retries for now
            "-a", cmd,
        ]
        processwrapper.check_output(cmd)

    @Driver.check_active
    @step()
    def on(self):
        self._switch("on")

    @Driver.check_active
    @step()
    def off(self):
        self._switch("off")

    @Driver.check_active
    @step()
    def cycle(self):
        self.off()
        time.sleep(self.delay)
        self.on()

    @Driver.check_active
    def get(self):
        cmd = self.hub.command_prefix + [
            self.tool,
            "-l", self.hub.path,
            "-p", str(self.hub.index),
        ]
        output = processwrapper.check_output(cmd)
        for line in output.splitlines():
            if not line or not line.startswith(b' '):
                continue
            prefix, status = line.strip().split(b':', 1)
            if not prefix == b"Port %d" % self.hub.index:
                continue
            status = status.split()
            if b"power" in status:
                return True
            if b"off" in status:
                return False
        raise ExecutionError(f"Did not find port status in uhubctl output ({repr(output)})")


@target_factory.reg_driver
@attr.s(eq=False)
class PDUDaemonDriver(Driver, PowerResetMixin, PowerProtocol):
    """PDUDaemonDriver - Driver using a PDU port available via pdudaemon"""
    bindings = {"port": "PDUDaemonPort", }
    delay = attr.ib(default=5.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._requests = import_module('requests')
        self._host = None
        self._port = None

    def _build_url(self, cmd):
        res = f"http://{self._host}:{self._port}/power/control/{cmd}?hostname={self.port.pdu}&port={self.port.index}"  # pylint: disable=line-too-long
        if cmd == 'reboot':
            res += f"&delay={math.ceil(self.delay)}"
        return res

    def on_activate(self):
        self._host, self._port = proxymanager.get_host_and_port(self.port, default_port=16421)

    @Driver.check_active
    @step()
    def on(self):
        r = self._requests.get(self._build_url('on'))
        r.raise_for_status()

    @Driver.check_active
    @step()
    def off(self):
        r = self._requests.get(self._build_url('off'))
        r.raise_for_status()

    @Driver.check_active
    @step()
    def cycle(self):
        r = self._requests.get(self._build_url('reboot'))
        r.raise_for_status()

    @Driver.check_active
    def get(self):
        raise NotImplementedError("pdudaemon does not support retrieving the port's state")



from labgrid.resource.pyvisa import PyVISADevice
from labgrid.driver.pyvisadriver import PyVISADriver

import pytest

pytest.importorskip("pyvisa")


def test_pyvisa_resource(target):
    PyVISADevice(target, name=None, type="TCPIP", url="127.0.0.1")


def test_resource_driver(target, mocker):
    PyVISADevice(target, name=None, type="TCPIP", url="127.0.0.1")
    driver = PyVISADriver(target, name=None)

    mocker.patch("pyvisa.ResourceManager.open_resource", return_value=None)
    target.activate(driver)



#!/usr/bin/python3

import argparse
import asyncio
import logging
import os
import struct
import sys
from abc import ABC, abstractmethod

HEADER = struct.Struct("<H")


class Pipe(ABC):
    def __init__(self, in_fd, out_fd):
        self.frame = b""
        self.buffer = b""
        self.read_event = asyncio.Event()
        self.write_event = asyncio.Event()
        self.eof = False
        self.loop = asyncio.get_running_loop()
        self.in_fd = in_fd
        self.out_fd = out_fd
        self.eof_handler = self.set_eof

    def _handle_eof(self):
        self.eof_handler()

    def set_eof(self):
        self.eof = True
        self.read_event.set()
        self.write_event.set()

    async def stream(self):
        while not self.eof:
            self.read_event.clear()
            self.loop.add_reader(self.in_fd, self.read_in)
            while not self.frame and not self.eof:
                await self.read_event.wait()
            self.loop.remove_reader(self.in_fd)

            self.write_event.clear()
            self.loop.add_writer(self.out_fd, self.write_out)
            while (self.frame or self.buffer) and not self.eof:
                await self.write_event.wait()
            self.loop.remove_writer(self.out_fd)

    @abstractmethod
    def read_in(self):
        raise NotImplementedError("Must be implemented in derived class")

    @abstractmethod
    def write_out(self):
        raise NotImplementedError("Must be implemented in derived class")


class StreamToTapPipe(Pipe):
    def read_until(self, length):
        while len(self.buffer) < length:
            d = os.read(self.in_fd, length - len(self.buffer))
            if not d:
                raise BrokenPipeError()
            self.buffer += d

    def read_in(self):
        if not self.frame:
            try:
                self.read_until(HEADER.size)

                hdr = HEADER.unpack_from(self.buffer)
                self.read_until(HEADER.size + hdr[0])

                self.frame = self.buffer[HEADER.size :]
                self.buffer = b""
                logging.debug("Read %d bytes from pipe", len(self.frame))

            except BlockingIOError:
                return
            except BrokenPipeError:
                logging.debug("Read pipe is closed")
                self._handle_eof()

        self.read_event.set()

    def write_out(self):
        if self.frame:
            try:
                w = os.write(self.out_fd, self.frame)
                if not w:
                    raise BrokenPipeError()
                logging.debug("Wrote %d bytes to tap", w)
                self.frame = b""
            except BlockingIOError:
                return
            except BrokenPipeError:
                logging.debug("Tap is closed")
                self._handle_eof()

        self.write_event.set()


class TapToStreamPipe(Pipe):
    def read_in(self):
        if not self.frame:
            try:
                d = os.read(self.in_fd, 1522)
                if not d:
                    raise BrokenPipeError()
                logging.debug("Read %d bytes from tap", len(d))
                self.frame = d
            except BlockingIOError:
                return
            except BrokenPipeError:
                logging.debug("Tap is closed")
                self._handle_eof()

        self.read_event.set()

    def write_out(self):
        if not self.buffer and self.frame:
            self.buffer = HEADER.pack(len(self.frame)) + self.frame
            self.frame = b""

        while self.buffer:
            try:
                w = os.write(self.out_fd, self.buffer)
                if not w:
                    raise BrokenPipeError()
                self.buffer = self.buffer[w:]
                logging.debug("Wrote %d bytes to pipe", w)
            except BlockingIOError:
                return
            except BrokenPipeError:
                logging.debug("Write pipe is closed")
                self._handle_eof()

        self.write_event.set()


async def pipe_loop(tap_fd, out_fd, in_fd):
    os.set_blocking(tap_fd, False)
    os.set_blocking(out_fd, False)
    os.set_blocking(in_fd, False)

    def handle_eof():
        in_to_tap.set_eof()
        tap_to_out.set_eof()

    in_to_tap = StreamToTapPipe(in_fd, tap_fd)
    tap_to_out = TapToStreamPipe(tap_fd, out_fd)

    in_to_tap.eof_handler = handle_eof
    tap_to_out.eof_handler = handle_eof

    return await asyncio.gather(in_to_tap.stream(), tap_to_out.stream())


def main():
    parser = argparse.ArgumentParser("Stream tap to stdio")
    parser.add_argument("fd", type=int, help="Tap FD to stream to stdio")
    parser.add_argument("--verbose", "-v", action="count", default=-1, help="Increase verbosity")

    args = parser.parse_args()

    if args.verbose >= 1:
        level = logging.DEBUG
    elif args.verbose >= 0:
        level = logging.INFO
    else:
        level = logging.WARNING
    root = logging.getLogger()
    root.setLevel(level)

    fmt = logging.Formatter(f"{os.getpid()} - %(asctime)s - %(name)s - %(levelname)s - %(message)s")

    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(fmt)

    root.addHandler(handler)

    # Duplicate stdin and stdout to new file descriptors for dedicated use by
    # the stream.
    with os.fdopen(os.dup(sys.stdin.fileno())) as in_f, os.fdopen(os.dup(sys.stdout.fileno())) as out_f:
        # Replace stdin with devnull
        with open(os.devnull, "r+") as devnull:
            os.dup2(devnull.fileno(), sys.stdin.fileno())

        # Replace stdout with stderr
        os.dup2(sys.stderr.fileno(), sys.stdout.fileno())

        asyncio.run(pipe_loop(args.fd, out_f.fileno(), in_f.fileno()))


if __name__ == "__main__":
    sys.exit(main())



import pytest
import os.path
from time import sleep
from shutil import which

from labgrid.resource.udev import SigrokUSBDevice, SigrokUSBSerialDevice
from labgrid.resource.sigrok import SigrokDevice
from labgrid.driver.sigrokdriver import SigrokDriver, SigrokPowerDriver


pytestmark = pytest.mark.skipif(not which("sigrok-cli"),
                              reason="sigrok not available")


def test_sigrok_resource(target):
    r = SigrokUSBDevice(target, name=None, match={"sys_name": "1-12"}, driver='fx2lafw', channels="D0,D1")


def test_sigrok_driver(target):
    r = SigrokDevice(target, name=None, driver='demo', channels="D0")
    d = SigrokDriver(target, name=None)
    target.activate(d)


@pytest.mark.sigrokusb
def test_sigrok_usb_driver(target, tmpdir):
    r = SigrokUSBDevice(target, name=None, match={"ID_MODEL_ID": "3881", "ID_VENDOR_ID": "0925"}, driver='fx2lafw', channels="D0,D1")
    d = SigrokDriver(target, name=None)
    target.activate(d)
    record = tmpdir.join("output.sr")
    d.capture(record)
    sleep(5)
    samples = d.stop()
    assert os.path.getsize(record) > 0
    assert samples != None
    assert list(samples[0].keys()) == ['time', 'D0', 'D1']
    assert list(samples[-1].keys()) == ['time', 'D0', 'D1']

def test_sigrok_power_driver(target):
    r = SigrokUSBSerialDevice(target, name=None, driver='manson-hcs-3xxx')
    r.avail = True
    d = SigrokPowerDriver(target, name=None)
    target.activate(d)



import contextlib
import errno
import os
import socket
import subprocess
import time

import pytest

from labgrid.util.agentwrapper import AgentWrapper
from labgrid.util.netns import NetNamespace


def netns_allowed():
    result = subprocess.run(["unshare", "-Umn", "true"])
    return result.returncode == 0


pytestmark = pytest.mark.skipif(not netns_allowed(), reason="missing user+mount+network namespace privileges")


FAMILY_PARAMS = (
    "family,host",
    [
        pytest.param(socket.AF_INET, "127.0.0.1", id="IPv4"),
        pytest.param(socket.AF_INET6, "::1", id="IPv6"),
    ],
)


@pytest.fixture
def netns():
    with AgentWrapper(None) as aw:
        netns = NetNamespace.create(aw, "be:df:8f:7a:12:db")

        links = netns.get_links()
        link_names = [link["ifname"] for link in links]
        assert netns.intf in link_names

        # Bring up lo in the namespace
        netns.run(["ip", "link", "set", "up", "dev", "lo"], check=True)

        yield netns

        # Verify that all sockets were closed in the namespace
        assert netns._agent.list_sockets() == []


@pytest.mark.parametrize(*FAMILY_PARAMS)
def test_tcp(netns, family, host):
    with netns.socket(family, socket.SOCK_STREAM) as server_sock:
        server_sock.bind((host, 5000))
        server_sock.listen(0)

        with netns.socket(family, socket.SOCK_STREAM) as client_sock:
            client_sock.connect((host, 5000))

            c, addr = server_sock.accept()
            client_sock.send(b"Hello")
            assert c.recv(1024) == b"Hello"

            c.send(b"World")
            assert client_sock.recv(1024) == b"World"


@pytest.mark.parametrize(*FAMILY_PARAMS)
def test_udp(netns, family, host):
    with netns.socket(family, socket.SOCK_DGRAM) as server_sock:
        server_sock.bind((host, 5000))
        assert server_sock.family == family

        # Test connected UDP client socket
        with netns.socket(family, socket.SOCK_DGRAM) as client_sock:
            client_sock.connect((host, 5000))
            client_sock.send(b"Hello")
            data, addr = server_sock.recvfrom(1024)
            assert data == b"Hello"

            server_sock.sendto(b"World", addr)
            data, addr = client_sock.recvfrom(1024)
            assert data == b"World"

        # Test unconnected UDP client socket
        server_addr = server_sock.getsockname()

        with netns.socket(family, socket.SOCK_DGRAM) as client_sock:
            client_sock.sendto(b"Hello", server_addr)
            data, addr = server_sock.recvfrom(1024)
            assert data == b"Hello"

            server_sock.sendto(b"World", addr)
            data, addr = client_sock.recvfrom(1024)
            assert data == b"World"


@pytest.mark.parametrize(*FAMILY_PARAMS)
def test_getaddrinfo(netns, family, host):
    # Test getaddrinfo
    for fam, _socktype, _proto, _cannonname, _sockaddr in netns.getaddrinfo(host, 5000):
        assert fam == family


def test_closed_socket(netns):
    # Test that closed sockets result in EBADF
    s = netns.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.close()
    with pytest.raises(OSError, check=lambda e: e.errno == errno.EBADF):
        s.bind(("127.0.0.1", 5000))


def test_dup(netns):
    with netns.socket(socket.AF_INET, socket.SOCK_DGRAM) as server_sock:
        server_sock.bind(("127.0.0.1", 5000))

        # Create, then close the duplicate socket. The server socket should still function
        dup_sock = server_sock.dup()
        dup_sock.close()

        with netns.socket(socket.AF_INET, socket.SOCK_DGRAM) as client_sock:
            client_sock.connect(("127.0.0.1", 5000))
            client_sock.send(b"Hello")
            data, addr = server_sock.recvfrom(1024)
            assert data == b"Hello"

            server_sock.sendto(b"World", addr)
            data, addr = client_sock.recvfrom(1024)
            assert data == b"World"


def test_detach(netns):
    with netns.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        fd = s.detach()
    os.close(fd)

    # Verify that all sockets were removed in the namespace
    assert netns._agent.list_sockets() == []


def test_socks(netns, tmp_path):
    with contextlib.ExitStack() as stack:
        (port, password) = stack.enter_context(netns.socks_proxy())

        import requests

        proxy_url = f"socks5h://labgrid:{password}@127.0.0.1:{port}"

        proxies = {"http": proxy_url, "https": proxy_url}

        session = requests.Session()
        session.proxies.update(proxies)

        tmp_file = tmp_path / "data"
        tmp_file.write_bytes(b"test\n")

        httpd = stack.enter_context(netns.Popen(["busybox", "httpd", "-f", "-h", str(tmp_path)]))
        stack.callback(lambda: httpd.terminate())

        time.sleep(1)

        response = session.get("http://127.0.0.1/data", timeout=5)
        assert response.content == b"test\n"
        assert response.status_code == 200

        response = session.get("http://127.0.0.1/missing", timeout=5)
        assert response.status_code == 404



import threading
from time import monotonic

import attr

from .common import ManagedResource, ResourceManager
from ..factory import target_factory


@attr.s(eq=False)
class MQTTManager(ResourceManager):
    _available = attr.ib(default=attr.Factory(set), validator=attr.validators.instance_of(set))
    _avail_lock = attr.ib(default=threading.Lock())
    _clients = attr.ib(default=attr.Factory(dict), validator=attr.validators.instance_of(dict))
    _topics = attr.ib(default=attr.Factory(list), validator=attr.validators.instance_of(list))
    _topic_lock = attr.ib(default=threading.Lock())
    _last = attr.ib(default=0.0, validator=attr.validators.instance_of(float))

    def _create_mqtt_connection(self, host, username, password):
        import paho.mqtt.client as mqtt

        client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)
        client.username_pw_set(username, password)
        client.connect(host)
        client.on_message = self._on_message
        client.loop_start()
        return client

    def on_resource_added(self, resource):
        host = resource.host
        username = resource.username
        password = resource.password
        key = f"{host}:{username or ''}:{password or ''}"
        if key not in self._clients:
            self._clients[key] = self._create_mqtt_connection(host, username, password)
        self._clients[key].subscribe(resource.avail_topic)

    def _on_message(self, client, userdata, msg):
        payload = msg.payload.decode("utf-8")
        topic = msg.topic
        if payload.lower() == "online":
            with self._avail_lock:
                self._available.add(topic)
        elif payload.lower() == "offline":
            with self._avail_lock:
                self._available.discard(topic)

    def poll(self):
        if monotonic() - self._last < 2:
            return  # ratelimit requests
        self._last = monotonic()
        with self._avail_lock:
            for resource in self.resources:
                resource.avail = resource.avail_topic in self._available


@target_factory.reg_resource
@attr.s(eq=False)
class MQTTResource(ManagedResource):
    manager_cls = MQTTManager

    host = attr.ib(validator=attr.validators.instance_of(str))
    avail_topic = attr.ib(validator=attr.validators.instance_of(str))
    username = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    password = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))

    def __attrs_post_init__(self):
        self.timeout = 30.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class TasmotaPowerPort(MQTTResource):
    power_topic = attr.ib(default=None, validator=attr.validators.instance_of(str))
    status_topic = attr.ib(default=None, validator=attr.validators.instance_of(str))



import abc


class VideoProtocol(abc.ABC):
    @abc.abstractmethod
    def get_qualities(self):
        raise NotImplementedError

    @abc.abstractmethod
    def stream(self, quality_hint=None):
        raise NotImplementedError



import abc


class InfoProtocol(abc.ABC):
    """Abstract class providing the InfoProtocol interface"""

    @abc.abstractmethod
    def get_ip(self, interface: str = "eth0"):
        """Implementations should return the IP address for the supplied
        interface."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_hostname(self):
        """Implementations should return the hostname for the supplied
        interface."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_service_status(self, service):
        """Implementations should return the status of a service"""
        raise NotImplementedError



"""
Driver for Digital Loggers PDU that use the REST API.
Tested with Ethernet Power Controller 7.

Based on https://www.digital-loggers.com/restapi.pdf

By default, only an authenticated user is allowed by REST API.

NetworkPowerPort:
    model: 'digitalloggers_restapi'
    host: 'http://admin:1234@192.168.0.100'
    index: 0
"""

from urllib.parse import urlparse

import requests
from requests.auth import HTTPDigestAuth
from requests.packages import urllib3


def extract_user_password_from_host(host):
    url = urlparse(host)
    if "@" in url.netloc:
        user = url.username
        password = url.password
        _host = f"{url.scheme}://{url.netloc.split('@')[1]}"
    else:
        user = None
        password = None
        _host = f"{url.scheme}://{url.netloc}"
    return user, password, _host


def power_set(host, port, index, value):
    # curl -u admin:1234 -v -X PUT -H "X-CSRF: x" --data 'value=true' --digest http://192.168.0.100/restapi/relay/outlets/=0/state/
    # curl -u admin:1234 -v -X PUT -H "X-CSRF: x" --data 'value=false' --digest http://192.168.0.100/restapi/relay/outlets/=0/state/
    # curl --insecure -u admin:1234 -v -X PUT -H "X-CSRF: x" --data 'value=true' --digest https://192.168.0.100/restapi/relay/outlets/=0/state/
    # curl --insecure -u admin:1234 -v -X PUT -H "X-CSRF: x" --data 'value=false' --digest https://192.168.0.100/restapi/relay/outlets/=0/state/
    assert port is None

    index = int(index)
    value = "true" if value else "false"
    payload = {"value": value}
    headers = {"X-CSRF": "x", "Accept": "application/json"}
    user, password, url = extract_user_password_from_host(host)
    host = f"{url}/restapi/relay/outlets/={index}/state/"
    with urllib3.warnings.catch_warnings():
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        if user and password:
            r = requests.put(host, data=payload, auth=HTTPDigestAuth(user, password), headers=headers, verify=False)
        else:
            r = requests.put(host, data=payload, headers=headers, verify=False)
        r.raise_for_status()


def power_get(host, port, index):
    # curl -u admin:1234 -v -X GET -H "X-CSRF: x" --digest http://192.168.0.100/restapi/relay/outlets/=0/state/
    # curl --insecure -u admin:1234 -v -X GET -H "X-CSRF: x" --digest https://192.168.0.100/restapi/relay/outlets/=0/state/
    assert port is None

    index = int(index)
    user, password, url = extract_user_password_from_host(host)
    headers = {"X-CSRF": "x", "Accept": "application/json"}
    host = f"{url}/restapi/relay/outlets/={index}/state/"
    with urllib3.warnings.catch_warnings():
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        if user and password:
            r = requests.get(host, auth=HTTPDigestAuth(user, password), headers=headers, verify=False)
        else:
            r = requests.get(host, headers=headers, verify=False)
        r.raise_for_status()
    statuses = r.json()
    return statuses[0]



"""A driver to control the Tinycontrol tcPDU
Reference: https://docs.tinycontrol.pl/en/tcpdu/api/commands

Example configuration to use port #3 on a device with URL 'http://172.17.180.53/'

NetworkPowerPort:
  model: tinycontrol_tcpdu
  host: 'http://172.17.180.53'
  index: 3
"""

from urllib.parse import urljoin

import requests


def power_set(host, port, index, value):
    assert port is None

    index = int(index)
    value = 1 if value else 0
    r = requests.get(urljoin(host, f"/api/v1/save/?out{index}={value}"))
    r.raise_for_status()


def power_get(host, port, index):
    assert port is None

    index = int(index)
    r = requests.get(urljoin(host, "/api/v1/read/status/?outValues"))
    r.raise_for_status()
    json_decoded = r.json()
    return json_decoded[f"out{index}"] == 1



from .base import SerialPort, NetworkInterface, EthernetPort, SysfsGPIO
from .ethernetport import SNMPEthernetPort
from .serialport import RawSerialPort, NetworkSerialPort
from .modbus import ModbusTCPCoil, WaveshareModbusTCPCoil
from .modbusrtu import ModbusRTU
from .networkservice import NetworkService
from .onewireport import OneWirePIO
from .power import NetworkPowerPort, PDUDaemonPort
from .remote import RemotePlace
from .udev import (
    AlteraUSBBlaster,
    AndroidUSBFastboot,
    DFUDevice,
    DeditecRelais8,
    HIDRelay,
    IMXUSBLoader,
    LXAUSBMux,
    MatchedSysfsGPIO,
    MXSUSBLoader,
    RKUSBLoader,
    SiSPMPowerPort,
    SigrokUSBDevice,
    SigrokUSBSerialDevice,
    USBAudioInput,
    USBDebugger,
    USBFlashableDevice,
    USBMassStorage,
    USBNetworkInterface,
    USBPowerPort,
    USBSDMuxDevice,
    USBSDWireDevice,
    USBSDWire3Device,
    USBSerialPort,
    USBTMC,
    USBVideo,
)
from .common import Resource, ResourceManager, ManagedResource
from .ykushpowerport import YKUSHPowerPort, NetworkYKUSHPowerPort
from .xenamanager import XenaManager
from .flashrom import Flashrom, NetworkFlashrom
from .docker import DockerManager, DockerDaemon, DockerConstants
from .lxaiobus import LXAIOBusPIO
from .pyvisa import PyVISADevice
from .provider import TFTPProvider, NFSProvider, HTTPProvider
from .mqtt import TasmotaPowerPort
from .httpvideostream import HTTPVideoStream
from .dediprogflasher import DediprogFlasher, NetworkDediprogFlasher
from .httpdigitalout import HttpDigitalOutput
from .sigrok import SigrokDevice
from .fastboot import AndroidNetFastboot
from .eth008 import Eth008DigitalOutput
from .laa import LAASerialPort, LAAPowerPort, LAAUSBGadgetMassStorage, \
                 LAAUSBPort, LAAButtonPort, \
                 LAALed, LAATempSensor, LAAWattMeter, LAAProvider



"""The SSHDriver uses SSH as a transport to implement CommandProtocol and FileTransferProtocol"""
import contextlib
import os
import re
import stat
import shlex
import shutil
import subprocess
import tempfile
import time
from functools import cached_property

import attr

from ..factory import target_factory
from ..protocol import CommandProtocol, FileTransferProtocol
from .commandmixin import CommandMixin
from .common import Driver
from ..step import step
from .exception import ExecutionError
from ..util.helper import get_free_port
from ..util.proxy import proxymanager
from ..util.timeout import Timeout
from ..util.ssh import get_ssh_connect_timeout


@target_factory.reg_driver
@attr.s(eq=False)
class SSHDriver(CommandMixin, Driver, CommandProtocol, FileTransferProtocol):
    """SSHDriver - Driver to execute commands via SSH"""
    bindings = {"networkservice": "NetworkService", }
    priorities = {CommandProtocol: 10, FileTransferProtocol: 10}
    keyfile = attr.ib(default="", validator=attr.validators.instance_of(str))
    stderr_merge = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    connection_timeout = attr.ib(default=float(get_ssh_connect_timeout()), validator=attr.validators.instance_of(float))
    explicit_sftp_mode = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    explicit_scp_mode = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    username = attr.ib(default="", validator=attr.validators.instance_of(str))
    password = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._keepalive = None
        self._ssh = self._get_tool("ssh")
        self._scp = self._get_tool("scp")
        self._sshfs = self._get_tool("sshfs")
        self._rsync = self._get_tool("rsync")

    def _get_tool(self, name):
        if self.target.env:
            return self.target.env.config.get_tool(name)
        return name

    def _get_username(self):
        """Get the username from this class or from NetworkService"""
        return self.username or self.networkservice.username

    def _get_password(self):
        """Get the password from this class or from NetworkService"""
        if self.password is not None:
            return self.password
        return self.networkservice.password

    def on_activate(self):
        self.ssh_prefix = ["-o", "LogLevel=ERROR"]
        if self.keyfile:
            keyfile_path = self.keyfile
            if self.target.env:
                keyfile_path = self.target.env.config.resolve_path(self.keyfile)
            self.ssh_prefix += ["-i", keyfile_path ]
        if self._get_password() is None:
            self.ssh_prefix += ["-o", "PasswordAuthentication=no"]

        self.control = self._start_own_master()
        self.ssh_prefix += ["-F", "none"]
        if self.control:
            self.ssh_prefix += ["-o", f"ControlPath={self.control.replace('%', '%%')}"]

        self._keepalive = None
        self._start_keepalive()

    def on_deactivate(self):
        try:
            self._stop_keepalive()
        finally:
            self._cleanup_own_master()

    @property
    def skip_deactivate_on_export(self):
        # We need to keep the connection to the target open.
        return True

    def _start_own_master(self):
        """Starts a controlmaster connection in a temporary directory."""

        timeout = Timeout(self.connection_timeout)

        # Retry start of controlmaster, to allow handle failures such as
        # connection refused during target startup
        connect_timeout = round(timeout.remaining)
        while True:
            try:
                return self._start_own_master_once(connect_timeout)
            except ExecutionError as e:
                if timeout.expired:
                    raise Exception("Timeout while waiting for ssh connection") from e
                time.sleep(0.5)
                connect_timeout = round(timeout.remaining)

    def _start_own_master_once(self, timeout):

        self.tmpdir = tempfile.mkdtemp(prefix='lg-ssh-')
        control = os.path.join(
            self.tmpdir, f'control-{self.networkservice.address}'
        )

        args = [self._ssh, "-f", *self.ssh_prefix, "-x", "-o", f"ConnectTimeout={timeout}",
                 "-o", "ControlPersist=300", "-o",
                 "UserKnownHostsFile=/dev/null", "-o", "StrictHostKeyChecking=no",
                 "-o", "ServerAliveInterval=15", "-MN", "-S", control.replace('%', '%%'), "-p",
                 str(self.networkservice.port), "-l", self._get_username(),
                 self.networkservice.address]

        # proxy via the exporter if we have an ifname suffix
        address = self.networkservice.address
        if address.count('%') > 1:
            raise ValueError(f"Multiple '%' found in '{address}'.")
        if '%' in address:
            address, ifname = address.split('%', 1)
        else:
            ifname = None

        proxy_cmd = proxymanager.get_command(self.networkservice, address, self.networkservice.port, ifname)
        if proxy_cmd:  # only proxy if needed
            args += [
                "-o", f"ProxyCommand={' '.join(proxy_cmd)} 2>{self.tmpdir}/proxy-stderr"
            ]

        env = os.environ.copy()
        pass_file = ''
        password = self._get_password()
        if password is not None:
            fd, pass_file = tempfile.mkstemp()
            os.fchmod(fd, stat.S_IRWXU)
            #with openssh>=8.4 SSH_ASKPASS_REQUIRE can be used to force SSH_ASK_PASS
            #openssh<8.4 requires the DISPLAY var and a detached process with start_new_session=True
            env = {'SSH_ASKPASS': pass_file, 'DISPLAY':'', 'SSH_ASKPASS_REQUIRE':'force'}
            with open(fd, 'w') as f:
                f.write("#!/bin/sh\necho " + shlex.quote(password))

        self.process = subprocess.Popen(args, env=env,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT,
                                        stdin=subprocess.DEVNULL,
                                        start_new_session=True)

        try:
            subprocess_timeout = timeout + 5
            return_value = self.process.wait(timeout=subprocess_timeout)
            if return_value != 0:
                stdout, _ = self.process.communicate(timeout=subprocess_timeout)
                stdout = stdout.split(b"\n")
                for line in stdout:
                    self.logger.warning("ssh: %s", line.rstrip().decode(encoding="utf-8", errors="replace"))

                try:
                    with open(f'{self.tmpdir}/proxy-stderr') as proxy_err_fd:
                        proxy_error = proxy_err_fd.read().strip()
                    if proxy_error:
                        raise ExecutionError(
                            f"Failed to connect to {self.networkservice.address} with {' '.join(args)}: error from SSH ProxyCommand: {proxy_error}",  # pylint: disable=line-too-long
                            stdout=stdout,
                        )
                except FileNotFoundError:
                    pass

                raise ExecutionError(
                    f"Failed to connect to {self.networkservice.address} with {' '.join(args)}: return code {return_value}",  # pylint: disable=line-too-long
                    stdout=stdout,
                )
        except subprocess.TimeoutExpired:
            raise ExecutionError(
                f"Subprocess timed out [{subprocess_timeout}s] while executing {args}",
            )
        finally:
            if self._get_password() is not None and os.path.exists(pass_file):
                os.remove(pass_file)

        if not os.path.exists(control):
            raise ExecutionError(
                f"no control socket to {self.networkservice.address}"
            )

        self.logger.info('Connected to %s', self.networkservice.address)

        return control

    @Driver.check_active
    @step(args=['cmd'], result=True)
    def run(self, cmd, codec="utf-8", decodeerrors="strict", timeout=None):
        return self._run(cmd, codec=codec, decodeerrors=decodeerrors, timeout=timeout)

    def _run(self, cmd, codec="utf-8", decodeerrors="strict", timeout=None):
        """Execute `cmd` on the target.

        This method runs the specified `cmd` as a command on its target.
        It uses the ssh shell command to run the command and parses the exitcode.
        cmd - command to be run on the target

        returns:
        (stdout, stderr, returncode)
        """
        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        complete_cmd = [self._ssh, "-x", *self.ssh_prefix,
                        "-p", str(self.networkservice.port), "-l", self._get_username(),
                        self.networkservice.address
                        ] + cmd.split(" ")
        self.logger.debug("Sending command: %s", complete_cmd)
        if self.stderr_merge:
            stderr_pipe = subprocess.STDOUT
        else:
            stderr_pipe = subprocess.PIPE
        try:
            sub = subprocess.Popen(
                complete_cmd, stdin=subprocess.DEVNULL, stdout=subprocess.PIPE, stderr=stderr_pipe
            )
        except:
            raise ExecutionError(
                f"error executing command: {complete_cmd}"
            )

        stdout, stderr = sub.communicate(timeout=timeout)
        stdout = stdout.decode(codec, decodeerrors).split('\n')
        if stdout[-1] == '':
            stdout.pop()
        if stderr is None:
            stderr = []
        else:
            stderr = stderr.decode(codec, decodeerrors).split('\n')
            stderr.pop()
        return (stdout, stderr, sub.returncode)

    def interact(self, cmd=None):
        assert cmd is None or isinstance(cmd, list)

        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        complete_cmd = [self._ssh, "-x", *self.ssh_prefix,
                        "-t",
                        self.networkservice.address
                        ]
        if cmd:
            complete_cmd += ["--", *cmd]
        self.logger.debug("Running command: %s", complete_cmd)
        sub = subprocess.Popen(
            complete_cmd,
        )
        return sub.wait()

    @contextlib.contextmanager
    def _forward(self, forward):
        cmd = [self._ssh, *self.ssh_prefix,
               "-O", "forward", forward,
               self.networkservice.address
               ]
        self.logger.debug("Running command: %s", cmd)
        subprocess.run(cmd, check=True)
        try:
            yield
        finally:
            cmd = [self._ssh, *self.ssh_prefix,
                   "-O", "cancel", forward,
                   self.networkservice.address
                   ]
            self.logger.debug("Running command: %s", cmd)
            # Master socket may have been cleaned up already, so don't bother
            # the user with an error message
            subprocess.run(cmd, stderr=subprocess.DEVNULL)

    @Driver.check_active
    @contextlib.contextmanager
    def forward_local_port(self, remoteport, localport=None):
        """Forward a local port to a remote port on the target

        A context manager that keeps a local port forwarded to a remote port as
        long as the context remains valid. A connection can be made to the
        returned port on localhost and it will be forwarded to the remote port
        on the target device

        usage:
            with ssh.forward_local_port(8080) as localport:
                # Use localhost:localport here to connect to port 8080 on the
                # target

        returns:
        localport
        """
        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        if localport is None:
            localport = get_free_port()

        forward = f"-L{localport:d}:localhost:{remoteport:d}"
        with self._forward(forward):
            yield localport

    @Driver.check_active
    @contextlib.contextmanager
    def forward_remote_port(self, remoteport, localport):
        """Forward a remote port on the target to a local port

        A context manager that keeps a remote port forwarded to a local port as
        long as the context remains valid. A connection can be made to the
        remote on the target device will be forwarded to the returned local
        port on localhost

        usage:
            with ssh.forward_remote_port(8080, 8081) as localport:
                # Connections to port 8080 on the target will be redirected to
                # localhost:8081
        """
        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        forward = f"-R{remoteport:d}:localhost:{localport:d}"
        with self._forward(forward):
            yield

    @Driver.check_active
    @contextlib.contextmanager
    def forward_unix_socket(self, unixsocket, localport=None):
        """Forward a unix socket on the target to a local port

        A context manager that keeps a unix socket forwarded to a local port as
        long as the context remains valid. A connection can be made to the
        remote socket on the target device will be forwarded to the returned
        local port on localhost

        usage:
            with ssh.forward_unix_socket("/run/docker.sock") as localport:
                # Use localhost:localport here to connect to the socket on the
                # target

        returns:
        localport
        """
        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        if localport is None:
            localport = get_free_port()

        forward = f"-L{localport:d}:{unixsocket:s}"
        with self._forward(forward):
            yield localport

    @Driver.check_active
    @step(args=['src', 'dst'])
    def scp(self, *, src, dst):
        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        if src.startswith(':') == dst.startswith(':'):
            raise ValueError("Either source or destination must be remote (start with :)")
        if src.startswith(':'):
            src = '_' + src
        if dst.startswith(':'):
            dst = '_' + dst

        complete_cmd = [self._scp,
                "-S", self._ssh,
                "-F", "none",
                "-o", f"ControlPath={self.control.replace('%', '%%')}",
                src, dst,
        ]
        
        if self.explicit_sftp_mode and self._scp_supports_explicit_sftp_mode():
            complete_cmd.insert(1, "-s")
        if self.explicit_scp_mode and self._scp_supports_explicit_scp_mode():
            complete_cmd.insert(1, "-O")

        self.logger.info("Running command: %s", complete_cmd)
        sub = subprocess.Popen(
            complete_cmd,
        )
        return sub.wait()

    @Driver.check_active
    @step(args=['src', 'dst', 'extra'])
    def rsync(self, *, src, dst, extra=[]):
        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        if src.startswith(':') == dst.startswith(':'):
            raise ValueError("Either source or destination must be remote (start with :)")
        if src.startswith(':'):
            src = '_' + src
        if dst.startswith(':'):
            dst = '_' + dst

        ssh_cmd = [self._ssh,
                "-F", "none",
                "-o", f"ControlPath={self.control.replace('%', '%%')}",
        ]

        complete_cmd = [self._rsync,
                "-v",
                f"--rsh={' '.join(ssh_cmd)}",
                "-rlpt",  # --recursive --links --perms --times
                "--one-file-system",
                "--progress",
                *extra,
                src, dst,
        ]
        self.logger.info("Running command: %s", complete_cmd)
        sub = subprocess.Popen(
            complete_cmd,
        )
        return sub.wait()

    @Driver.check_active
    @step(args=['path', 'mountpoint'])
    def sshfs(self, *, path, mountpoint):
        if not self._check_keepalive():
            raise ExecutionError("Keepalive no longer running")

        complete_cmd = [self._sshfs,
                "-F", "none",
                "-f",
                "-o", f"ControlPath={self.control.replace('%', '%%')}",
                f":{path}",
                mountpoint,
        ]

        self.logger.debug("Running command: %s", complete_cmd)
        sub = subprocess.Popen(
            complete_cmd,
        )
        try:
            sub.wait(1)
            raise ExecutionError(
                f"error executing command: {complete_cmd}"
            )
        except subprocess.TimeoutExpired:  # still running
            self.logger.info("Started SSHFS on %s. Press CTRL-C to stop.", mountpoint)

        sub.wait()

    def get_status(self):
        """The SSHDriver is always connected, return 1"""
        return 1

    @cached_property
    def _ssh_version(self):
        version = subprocess.run([self._ssh, "-V"], capture_output=True, text=True)
        version = re.match(r"^OpenSSH_(\d+)\.(\d+)", version.stderr)
        return tuple(int(x) for x in version.groups())

    def _scp_supports_explicit_sftp_mode(self):
        major, minor = self._ssh_version

        # OpenSSH >= 8.6 supports explicitly using the SFTP protocol via -s
        if major == 8 and minor >= 6:
            return True
        # OpenSSH >= 9.0 default to the SFTP protocol
        if major >= 9:
            return False
        raise Exception(f"OpenSSH version {major}.{minor} does not support explicit SFTP mode")

    def _scp_supports_explicit_scp_mode(self):
        major, minor = self._ssh_version

        # OpenSSH >= 9.0 default to the SFTP protocol
        if major >= 9:
            return True
        raise Exception(f"OpenSSH version {major}.{minor} does not support explicit SCP mode")

    @Driver.check_active
    @step(args=['filename', 'remotepath'])
    def put(self, filename, remotepath=''):
        transfer_cmd = [
            self._scp,
            "-S", self._ssh,
            *self.ssh_prefix,
            "-P", str(self.networkservice.port),
            "-r",
            filename,
            f"{self._get_username()}@{self.networkservice.address}:{remotepath}"
            ]

        if self.explicit_sftp_mode and self._scp_supports_explicit_sftp_mode():
            transfer_cmd.insert(1, "-s")
        if self.explicit_scp_mode and self._scp_supports_explicit_scp_mode():
            transfer_cmd.insert(1, "-O")

        try:
            sub = subprocess.call(
                transfer_cmd
            )  #, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        except:
            raise ExecutionError(
                f"error executing command: {transfer_cmd}"
            )
        if sub != 0:
            raise ExecutionError(
                f"error executing command: {transfer_cmd}"
            )

    @Driver.check_active
    @step(args=['filename', 'destination'])
    def get(self, filename, destination="."):
        transfer_cmd = [
            self._scp,
            "-S", self._ssh,
            *self.ssh_prefix,
            "-P", str(self.networkservice.port),
            "-r",
            f"{self._get_username()}@{self.networkservice.address}:{filename}",
            destination
            ]

        if self.explicit_sftp_mode and self._scp_supports_explicit_sftp_mode():
            transfer_cmd.insert(1, "-s")
        if self.explicit_scp_mode and self._scp_supports_explicit_scp_mode():
            transfer_cmd.insert(1, "-O")

        try:
            sub = subprocess.call(
                transfer_cmd
            )  #, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        except:
            raise ExecutionError(
                f"error executing command: {transfer_cmd}"
            )
        if sub != 0:
            raise ExecutionError(
                f"error executing command: {transfer_cmd}"
            )

    def _cleanup_own_master(self):
        """Exit the controlmaster and delete the tmpdir"""
        complete_cmd = f"{self._ssh} -x -o ControlPath={self.control.replace('%', '%%')} -O exit -p {self.networkservice.port} -l {self._get_username()} {self.networkservice.address}".split(' ')  # pylint: disable=line-too-long
        res = subprocess.call(
            complete_cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )

        if res != 0:
            self.logger.info("Socket already closed")

        self.process.communicate()

        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def _start_keepalive(self):
        """Starts a keepalive connection via the own or external master."""
        args = [self._ssh, *self.ssh_prefix, self.networkservice.address, "cat"]

        assert self._keepalive is None
        self._keepalive = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
        )

        self.logger.debug('Started keepalive for %s', self.networkservice.address)

    def _check_keepalive(self):
        return self._keepalive.poll() is None

    def _stop_keepalive(self):
        assert self._keepalive is not None

        self.logger.debug('Stopping keepalive for %s', self.networkservice.address)

        stdout = None
        try:
            stdout, _ = self._keepalive.communicate(timeout=60)
        except subprocess.TimeoutExpired:
            self._keepalive.kill()
            try:
                # Try again to get output
                stdout, _ = self._keepalive.communicate(timeout=60)
            except subprocess.TimeoutExpired:
                self.logger.warning("ssh keepalive for %s timed out during termination", self.networkservice.address)
        finally:
            self._keepalive = None

        if stdout:
            for line in stdout.splitlines():
                self.logger.warning("Keepalive %s: %s", self.networkservice.address, line)



from labgrid.resource.remote import NetworkUSBVideo
from labgrid.driver.usbvideodriver import USBVideoDriver


def test_usbvideo_create(target):
    r = NetworkUSBVideo(
        target,
        name=None,
        host="localhost",
        busnum=0,
        devnum=1,
        path="0:1",
        vendor_id=0x0,
        model_id=0x0,
    )
    d = USBVideoDriver(target, name=None)
    assert isinstance(d, USBVideoDriver)



def test_barebox(command):
    stdout, stderr, returncode = command.run("version")
    assert returncode == 0
    assert stdout
    assert not stderr
    assert "barebox" in "\n".join(stdout)

    stdout, stderr, returncode = command.run("false")
    assert returncode == 1
    assert not stdout
    assert not stderr



"""Rest interface for controlling power port, using PUT / GET on a URL.

NetworkPowerPort:
    model: rest
    host: 'http://192.168.0.42/relay/{index}/value'
    index: 3

Will do a GET request to http://192.168.0.42/relay/3/value to get current
relay state, expecting a response of either '0' (relay off) or '1' (relay
on), and a PUT request to http://192.168.0.42/relay/3/value with request
data of '0' or '1' to change relay state.

"""

import requests


def power_set(host, port, index, value):
    assert port is None
    value = b"1" if value else b"0"
    r = requests.put(host.format(index=index), data=value)
    r.raise_for_status()


def power_get(host, port, index):
    assert port is None
    r = requests.get(host.format(index=index))
    r.raise_for_status()
    return r.text == "1"



import os
import re
import time

import pytest
import pexpect

def test_startup(coordinator):
    pass

@pytest.fixture(scope='function')
def place(create_place):
    create_place('test')

@pytest.fixture(scope='function')
def create_place(coordinator):
    place_names = []

    def _create_place(place_name):
        with pexpect.spawn(f'python -m labgrid.remote.client -p {place_name} create') as spawn:
            spawn.expect(pexpect.EOF)
        assert spawn.exitstatus == 0, spawn.before.strip()

        place_names.append(place_name)

        with pexpect.spawn(f'python -m labgrid.remote.client -p {place_name} set-tags board=123board') as spawn:
            spawn.expect(pexpect.EOF)
        assert spawn.exitstatus == 0, spawn.before.strip()

    yield _create_place

    for place_name in place_names:
        # clean up
        with pexpect.spawn(f'python -m labgrid.remote.client -p {place_name} release') as spawn:
            spawn.expect(pexpect.EOF)

        with pexpect.spawn(f'python -m labgrid.remote.client -p {place_name} delete') as spawn:
            spawn.expect(pexpect.EOF)
        assert spawn.exitstatus == 0, spawn.before.strip()

@pytest.fixture(scope='function')
def place_acquire(place, exporter):
    with pexpect.spawn('python -m labgrid.remote.client -p test add-match "*/Testport/*"') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    yield

    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_connect_error():
    with pexpect.spawn('python -m labgrid.remote.client -x 127.0.0.1:20409 places') as spawn:
        spawn.expect("Could not connect to coordinator")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 1, spawn.before.strip()

def test_connect_timeout(coordinator):
    coordinator.suspend_tree()
    try:
        with pexpect.spawn('python -m labgrid.remote.client places') as spawn:
            spawn.expect("connection attempt timed out before receiving SETTINGS frame")
            spawn.expect(pexpect.EOF)
            spawn.close()
            assert spawn.exitstatus == 1, spawn.before.strip()
    finally:
        coordinator.resume_tree()
        pass

def test_place_show(place):
    with pexpect.spawn('python -m labgrid.remote.client -p test show') as spawn:
        spawn.expect("Place 'test':")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_alias(place):
    with pexpect.spawn('python -m labgrid.remote.client -p test add-alias foo') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p foo del-alias foo') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_comment(place):
    with pexpect.spawn('python -m labgrid.remote.client -p test set-comment my comment') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test show') as spawn:
        spawn.expect("Place 'test':")
        spawn.expect(" comment: my comment")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_match(place):
    with pexpect.spawn('python -m labgrid.remote.client -p test add-match "e1/g1/r1" "e2/g2/*"') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test del-match "e1/g1/r1"') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test show') as spawn:
        spawn.expect(" matches:")
        spawn.expect(" e2/g2/*")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_match_duplicates(place):
    # first given match should succeed, second should be skipped
    matches = (
        ("e1/g1/r1", "e1/g1/r1"),
        ("e1/g1/r1/n1", "e1/g1/r1/n1"),
        ("e1/g1/r1/n1", "e1/g1/r1"),
        ("e1/g1/r1", "e1/g1/r1/n1"),
    )
    for match in matches:
        with pexpect.spawn(f'python -m labgrid.remote.client -p test add-match "{match[0]}" "{match[1]}"') as spawn:
            spawn.expect(f"pattern '{match[1]}' exists, skipping")
            spawn.expect(pexpect.EOF)
            spawn.close()
            assert spawn.exitstatus == 0, spawn.before.strip()

        with pexpect.spawn(f'python -m labgrid.remote.client -p test del-match "{match[0]}"') as spawn:
            spawn.expect(pexpect.EOF)
            spawn.close()
            assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_acquire(place):
    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client who') as spawn:
        spawn.expect(".*test.*")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_acquire_multiple(create_place, tmpdir):
    # create multiple places
    place_names = ['test1', 'test2']
    for place_name in place_names:
        create_place(place_name)

    # create env config with multiple RemotePlaces
    p = tmpdir.join('config.yaml')
    p.write('targets:')
    for place_name in place_names:
        p.write(
            f"""
        {place_name}:
          resources:
            RemotePlace:
              name: {place_name}
        """,
            mode='a',
        )

    # acquire all places in env config
    with pexpect.spawn(f'python -m labgrid.remote.client -c {p} acquire') as spawn:
        spawn.expect(pexpect.EOF)
    assert spawn.exitstatus == 0, spawn.before.strip()

    # check 'who'
    with pexpect.spawn('python -m labgrid.remote.client who') as spawn:
        spawn.expect(pexpect.EOF)
        for place_name in place_names:
            assert place_name.encode('utf-8') in spawn.before

    assert spawn.exitstatus == 0, spawn.before.strip()

    # release all places in env config
    with pexpect.spawn(f'python -m labgrid.remote.client -c {p} release') as spawn:
        spawn.expect(pexpect.EOF)
    assert spawn.exitstatus == 0, spawn.before.strip()

    # check 'who' again
    with pexpect.spawn('python -m labgrid.remote.client who') as spawn:
        spawn.expect('User.*Host.*Place.*Changed\r\n')
    assert not spawn.before, spawn.before

def test_place_acquire_enforce(place):
    with pexpect.spawn('python -m labgrid.remote.client -p test add-match does/not/exist') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect("Match does/not/exist has no matching remote resource")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus != 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test acquire --allow-unmatched') as spawn:
        spawn.expect("acquired place test")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_acquire_broken(place, exporter):
    with pexpect.spawn('python -m labgrid.remote.client -p test add-match "*/Broken/*"') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect('Failed to acquire resources for place test')
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 1, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test show') as spawn:
        spawn.expect("'broken': 'start failed'")
        spawn.expect(pexpect.EOF)
        spawn.close()
        print(spawn.before.decode())
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_place_release_from(monkeypatch, place, exporter):
    user = "test-user"
    host = "test-host"
    monkeypatch.setenv("LG_USERNAME", user)
    monkeypatch.setenv("LG_HOSTNAME", host)

    # Acquire place
    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    # Ensure place is acquired by user
    with pexpect.spawn('python -m labgrid.remote.client who') as spawn:
        spawn.expect(f"{user}\\s+{host}\\s+test")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    # Use release-from to release for a different user
    with pexpect.spawn('python -m labgrid.remote.client -p test release-from foo/bar') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    # Ensure place is still acquired by this user
    with pexpect.spawn('python -m labgrid.remote.client who') as spawn:
        spawn.expect(f"{user}\\s+{host}\\s+test")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    # Use release-from to release place for this user
    with pexpect.spawn(f'python -m labgrid.remote.client -p test release-from {host}/{user}') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    # Ensure place is still no longer acquired
    with pexpect.spawn('python -m labgrid.remote.client who') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        before = spawn.before.decode("utf-8").strip()
        assert user not in before and not host in before, before

def test_place_add_no_name(coordinator):
    with pexpect.spawn('python -m labgrid.remote.client create') as spawn:
        spawn.expect("missing place name")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus != 0, spawn.before.strip()

def test_place_del_no_name(coordinator):
    with pexpect.spawn('python -m labgrid.remote.client delete') as spawn:
        spawn.expect("place pattern not specified")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus != 0, spawn.before.strip()

def test_remoteplace_target(place_acquire, tmpdir):
    from labgrid.environment import Environment
    p = tmpdir.join("config.yaml")
    p.write(
        """
    targets:
      test1:
        role: foo
        resources:
          RemotePlace:
            name: test
    """
    )
    e = Environment(str(p))
    t = e.get_target("test1")
    t.await_resources(t.resources)

    remote_place = t.get_resource("RemotePlace")
    assert remote_place.tags == {"board": "123board"}

def test_remoteplace_target_without_env(request, place_acquire):
    from labgrid import Target
    from labgrid.resource import RemotePlace

    t = Target(request.node.name)
    remote_place = RemotePlace(t, name="test")
    assert remote_place.tags == {"board": "123board"}

def test_resource_conflict(place_acquire, tmpdir):
    with pexpect.spawn('python -m labgrid.remote.client -p test2 create') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test2 add-match "*/Testport/*"') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test2 acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus != 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test2 delete') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_reservation(place_acquire, tmpdir):
    with pexpect.spawn('python -m labgrid.remote.client reserve --shell board=123board name=test') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        m = re.search(rb"^export LG_TOKEN=(\S+)$", spawn.before.replace(b'\r\n', b'\n'), re.MULTILINE)
        assert m is not None, spawn.before.strip()
        token = m.group(1)

    env = os.environ.copy()
    env['LG_TOKEN'] = token.decode('ASCII')

    with pexpect.spawn('python -m labgrid.remote.client reservations') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b'waiting' in spawn.before, spawn.before.strip()
        assert token in spawn.before, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client reservations') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b'allocated' in spawn.before, spawn.before.strip()
        assert token in spawn.before, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client reservations') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b'allocated' in spawn.before, spawn.before.strip()
        assert token in spawn.before, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p + acquire', env=env) as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p + show', env=env) as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert token in spawn.before, spawn.before.strip()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p + release', env=env) as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client reservations') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b'allocated' in spawn.before, spawn.before.strip()
        assert token in spawn.before, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client cancel-reservation', env=env) as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client reservations') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert token not in spawn.before, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_resource_acquired_state_on_exporter_restart(monkeypatch, place, exporter):
    user = "test-user"
    host = "test-host"
    monkeypatch.setenv("LG_USERNAME", user)
    monkeypatch.setenv("LG_HOSTNAME", host)

    # add resource match
    with pexpect.spawn('python -m labgrid.remote.client -p test add-match testhost/Testport/NetworkSerialPort') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    # make sure matching resource is found
    with pexpect.spawn('python -m labgrid.remote.client -p test show') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b"acquired: None" in spawn.before
        assert b"Matching resource 'NetworkSerialPort' (testhost/Testport/NetworkSerialPort/NetworkSerialPort)" in spawn.before

    with pexpect.spawn('python -m labgrid.remote.client -p test -v resources') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b"Resource 'NetworkSerialPort' (testhost/Testport/NetworkSerialPort[/NetworkSerialPort]):\r\n      {'acquired': None," in spawn.before

    # lock place (and its resources)
    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test -v resources') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b"Resource 'NetworkSerialPort' (testhost/Testport/NetworkSerialPort[/NetworkSerialPort]):\r\n      {'acquired': 'test'," in spawn.before

    # restart exporter
    exporter.stop()
    exporter.start()

    # make sure matching resource is still found
    with pexpect.spawn('python -m labgrid.remote.client -p test show') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert f"acquired: {host}/{user}" in spawn.before.decode("utf-8")
        assert b"Acquired resource 'NetworkSerialPort' (testhost/Testport/NetworkSerialPort/NetworkSerialPort)" in spawn.before

    # release place
    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test -v resources') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b"Resource 'NetworkSerialPort' (testhost/Testport/NetworkSerialPort[/NetworkSerialPort]):\r\n      {'acquired': None," in spawn.before

    # make sure matching resource is still found
    with pexpect.spawn('python -m labgrid.remote.client -p test show') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b"acquired: None" in spawn.before
        assert b"Matching resource 'NetworkSerialPort' (testhost/Testport/NetworkSerialPort/NetworkSerialPort)" in spawn.before

    # place should now be acquirable again
    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()


def test_exporter_timeout(place, exporter):
    with pexpect.spawn('python -m labgrid.remote.client resources') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert b'/Testport/NetworkSerialPort' in spawn.before

    # lock resources ensure cleanup is needed
    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    exporter.suspend_tree()
    try:
        time.sleep(30)

        # the unresponsive exporter should be kicked by now
        with pexpect.spawn('python -m labgrid.remote.client resources') as spawn:
            spawn.expect(pexpect.EOF)
            spawn.close()
            assert spawn.exitstatus == 0, spawn.before.strip()
            assert b'/Testport/NetworkSerialPort' not in spawn.before
    finally:
        exporter.resume_tree()

    # the exporter should quit by itself now
    time.sleep(5)

    assert not exporter.isalive()
    assert exporter.exitstatus == 100

    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_reservation_custom_config(place, exporter, tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
    """
    targets:
      test1:
        role: foo
        resources:
          RemotePlace:
            name: test
    """
    )
    with pexpect.spawn(f'python -m labgrid.remote.client -c {p} reserve --wait --shell board=123board name=test') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        m = re.search(rb"^export LG_TOKEN=(\S+)$", spawn.before.replace(b'\r\n', b'\n'), re.MULTILINE)
        s = re.search(rb"^Selected role$", spawn.before.replace(b'\r\n', b'\n'), re.MULTILINE)
        assert m is not None, spawn.before.strip()
        assert s is None, spawn.before.strip()
        token = m.group(1)

    env = os.environ.copy()
    env['LG_TOKEN'] = token.decode('ASCII')

    with pexpect.spawn(f'python -m labgrid.remote.client -c {p} -p + lock', env=env) as spawn:
        spawn.expect("acquired place test")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn(f'python -m labgrid.remote.client -c {p} -p + release', env=env) as spawn:
        spawn.expect("released place test")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

def test_same_name_resources(place, exporter, tmpdir):
    with pexpect.spawn('python -m labgrid.remote.client -p test add-named-match "testhost/Many/NetworkService" "samename"') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test add-named-match "testhost/Many/NetworkSerialPort" "samename"') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test acquire') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test env') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()
        assert "NetworkService".encode("utf-8") in spawn.before.replace(b'\r\n', b'\n'), spawn.before.strip()
        assert "NetworkSerialPort".encode("utf-8") in spawn.before.replace(b'\r\n', b'\n'), spawn.before.strip()

    with pexpect.spawn('python -m labgrid.remote.client -p test release') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before.strip()



"""A driver to control the Tinycontrol IP Power Socket 6G10A v2
Reference: https://tinycontrol.pl/media/documents/manual_IP_Power_Socket__6G10A_v2_LANLIS-010-015_En-1.pdf

Example configuration to use port #3 on a device with URL 'http://172.17.180.53:9999/'

NetworkPowerPort:
  model: tinycontrol
  host: 'http://172.17.180.53:9999/'
  index: 3
"""

from urllib.parse import urljoin
import xml.etree.ElementTree as ET

import requests


def power_set(host, port, index, value):
    assert port is None

    index = int(index)
    value = 1 if value else 0
    r = requests.get(urljoin(host, f"/outs.cgi?out{index}={value}"))
    r.raise_for_status()


def power_get(host, port, index):
    assert port is None

    index = int(index)
    r = requests.get(urljoin(host, "/st0.xml"))
    r.raise_for_status()
    root = ET.fromstring(r.text)
    output = root.find(f"out{index}")
    return output.text == "1"



import inspect

from .exceptions import InvalidConfigError, RegistrationError
from .util.dict import filter_dict


class TargetFactory:
    def __init__(self):
        self.resources = {}
        self.drivers = {}
        self.all_classes = {}

    def reg_resource(self, cls):
        """Register a resource with the factory.

        Returns the class to allow using it as a decorator."""
        cls_name = cls.__name__
        if cls_name in self.all_classes:
            raise RegistrationError(f"resource with name {cls_name} was already registered")
        self.resources[cls.__name__] = cls
        self._insert_into_all(cls)
        return cls

    def reg_driver(self, cls):
        """Register a driver with the factory.

        Returns the class to allow using it as a decorator."""
        cls_name = cls.__name__
        if cls_name in self.all_classes:
            raise RegistrationError(f"driver with name {cls_name} was already registered")
        self.drivers[cls_name] = cls
        self._insert_into_all(cls)
        return cls

    @staticmethod
    def _convert_to_named_list(data):
        """Convert a tree of resources or drivers to a named list.

        When using named resources or drivers, the config file uses a list of
        dicts instead of simply nested dicts. This allows creating multiple
        instances of the same class with different names.

        resources: # or drivers
          FooPort: {}
          BarPort:
            name: "bar"

        or

        resources: # or drivers
        - FooPort: {}
        - BarPort:
            name: "bar"

        should be transformed to

        resources: # or drivers
        - cls: "FooPort"
        - cls: "BarPort"
          name: "bar"
        """

        # resolve syntactic sugar (list of dicts each containing a dict of key -> args)
        result = []
        if isinstance(data, list):
            for item in data:
                if not isinstance(item, dict):
                    raise InvalidConfigError(
                        f"invalid list item type {type(item)} (should be dict)")
                if not item:
                    raise InvalidConfigError("invalid empty dict as list item")
                if len(item) > 1:
                    if 'cls' in item:
                        item = item.copy()
                    else:
                        raise InvalidConfigError(f"missing 'cls' key in {item}")
                else:
                    # only one pair left
                    (key, value), = item.items()
                    if key == 'cls':
                        item = item.copy()
                    else:
                        item = {'cls':  key}
                        if value is None:
                            raise InvalidConfigError("invalid list item, add empty dict for no arguments")  # pylint: disable=line-too-long
                        item.update(value)
                result.append(item)
        elif isinstance(data, dict):
            for cls, args in data.items():
                args = args.copy()
                args.setdefault('cls', cls)
                result.append(args)
        else:
            raise InvalidConfigError(f"invalid type {type(data)} (should be dict or list)")
        for item in result:
            item.setdefault('name', None)
            assert 'cls' in item
        return result

    @staticmethod
    def normalize_config(config):
        resources = {}
        drivers = {}
        for item in TargetFactory._convert_to_named_list(config.get('resources', {})):
            resource = item.pop('cls')
            name = item.pop('name', None)
            args = item # remaining args
            resources.setdefault(resource, {})[name] = (args, )
        for item in TargetFactory._convert_to_named_list(config.get('drivers', {})):
            driver = item.pop('cls')
            name = item.pop('name', None)
            bindings = item.pop('bindings', {})
            args = item # remaining args
            drivers.setdefault(driver, {})[name] = (args, bindings)
        return resources, drivers

    def make_resource(self, target, resource, name, args):
        assert isinstance(args, dict)
        if not resource in self.resources:
            raise InvalidConfigError(f"unknown resource class {resource}")
        try:
            cls = self.resources[resource]
            args = filter_dict(args, cls, warn=True)
            r = cls(target, name, **args)
        except TypeError as e:
            raise InvalidConfigError(
                f"failed to create {resource} for target '{target}' using {args} "
            ) from e
        return r

    def make_driver(self, target, driver, name, args):
        assert isinstance(args, dict)
        if not driver in self.drivers:
            raise InvalidConfigError(f"unknown driver class {driver}")
        try:
            cls = self.drivers[driver]
            args = filter_dict(args, cls, warn=True)
            d = cls(target, name, **args)
        except TypeError as e:
            raise InvalidConfigError(
                f"failed to create {driver} for target '{target}' using {args} ") from e
        return d

    def make_target(self, name, config, *, env=None):
        from .target import Target

        target = Target(name, env=env)
        for item in TargetFactory._convert_to_named_list(config.get('resources', {})):
            resource = item.pop('cls')
            name = item.pop('name', None)
            args = item # remaining args
            self.make_resource(target, resource, name, args)
        for item in TargetFactory._convert_to_named_list(config.get('drivers', {})):
            driver = item.pop('cls')
            name = item.pop('name', None)
            bindings = item.pop('bindings', {})
            args = item # remaining args
            target.set_binding_map(bindings)
            self.make_driver(target, driver, name, args)
        return target

    def class_from_string(self, string: str):
        try:
            return self.all_classes[string]
        except KeyError:
            raise KeyError(f"No driver/resource/protocol of type '{string}' in factory, perhaps not registered?")

    def _insert_into_all(self, cls):
        classes = inspect.getmro(cls)
        for cl in classes:
            if not self.all_classes.get(cl.__name__):
                self.all_classes[cl.__name__] = cl


#: Global TargetFactory instance
#:
#: This instance is used to register Resource and Driver classes so that
#: Targets can be created automatically from YAML files.
target_factory = TargetFactory()



""" tested with Cisco CBS350, should be compatible with switches implementing the PoE administration MiB"""

from ..exception import ExecutionError
from ...util.snmp import SimpleSNMP

OID = "1.3.6.1.2.1.105.1.1.1.3.1"

def power_set(host, port, index, value):
    _snmp = SimpleSNMP(host, 'private', port=port)
    outlet_control_oid = "{}.{}".format(OID, index)

    oid_value = "1" if value else "2"

    _snmp.set(outlet_control_oid, oid_value)

def power_get(host, port, index):
    _snmp = SimpleSNMP(host, 'private', port=port)
    output_status_oid = "{}.{}".format(OID, index)

    value = _snmp.get(output_status_oid)

    if value == 1:  # On
        return True
    if value == 2:  # Off
        return False

    raise ExecutionError("failed to get SNMP value")



import requests

from ..exception import ExecutionError

PORT = 80


def power_set(host, port, index, value):
    index = int(index)
    assert 1 <= index <= 8
    # access the web interface...
    value = 1 if value else 0
    r = requests.get(f"http://{host}:{port}/switch.html?cmd=1&p={index}&s={value}")
    r.raise_for_status()


def power_get(host, port, index):
    index = int(index)
    assert 1 <= index <= 8
    # get the contents of the main page
    r = requests.get(f"http://{host}:{port}/")
    r.raise_for_status()
    for line in r.text.splitlines():
        power_pattern = f"Power Port {index}</td>"
        switch_patern = f"SwitchPort {index}</td>"
        if line.find(power_pattern) > 0 or line.find(switch_patern) > 0:
            if line.find("OFF") > 0:
                return False
            if line.find("ON") > 0:
                return True

            raise ExecutionError("failed to parse the port status")
    # if we got this far, something is wrong with the website
    raise ExecutionError("failed to find the port")
